import click

from todotree.Config.ConsolePrefixes import ConsolePrefixes


class ConsolePrefixesInit(ConsolePrefixes):
    """
    Provides additional features to ConsolePrefixes.
    These are only used in the Init command.
    """
    def __init__(self, enable_colors: bool,
                 info_prefix: str, warning_prefix: str, error_prefix: str,
                 info_color: str, warn_color: str, error_color: str):
        super().__init__(enable_colors,
                         info_prefix, warning_prefix, error_prefix,
                         info_color, warn_color, error_color)
        self.enable_colors = enable_colors
        self.error_prefix = error_prefix
        self.warning_prefix = warning_prefix
        self.info_prefix = info_prefix
        self.info_color = info_color
        self.warn_color = warn_color
        self.error_color = error_color

    @staticmethod
    def from_console_prefixes(c: ConsolePrefixes):
        """Converts from console prefixes."""
        return ConsolePrefixesInit(
            enable_colors=c.enable_colors,
            info_prefix=c.info_prefix,
            warning_prefix=c.warning_prefix,
            error_prefix=c.error_prefix,
            info_color=c.info_color,
            warn_color=c.warn_color,
            error_color=c.error_color
        )

    def prompt_menu(self, question: str, answers: list, custom_answer: str | None = None) -> str | int:
        """
        Ask the end user for input using a menu.
        @param question: The question to ask the user.
        @param answers: The predefined answers that the user can select.
        @param custom_answer: The custom option Text. Do not append with a dot.
            If the user chooses the Custom answer option. The user will be prompted to fill in the custom answer.
        @return: The answer that the user chose.
        If the user chooses a predefined option, it will be number of that option.
        It will be a string if the answer came from the custom option.
        """
        confirmed = False
        while not confirmed:
            text = question
            click.echo(text)
            for i, answer in enumerate(answers):
                self._menu_line(i, answer)
            custom_answer_number = len(answers)
            if custom_answer:
                self._menu_line(custom_answer_number, custom_answer)

            choices = [str(i) for i in range(0, custom_answer_number + 1)]
            answer_int = int(self.prompt(text='', show_choices=False, type=click.Choice(choices),
                                         prompt_suffix="Your answer:"))
            if custom_answer and answer_int == custom_answer_number:
                answer = self.prompt(text=f"{custom_answer} option chosen. Please type it in.")
                confirmed = self.confirm(text=f"Is this correct?: {answer}")
            else:
                answer = answer_int
                confirmed = self.confirm(text=f"Is this correct?: {answers[answer_int]}")
        return answer

    def _menu_line(self, answer_number: int, answer: str):
        """Prints a menu line (with colors if enabled."""
        if self.enable_colors:
            click.secho(f"  [{str(answer_number)}] ", fg=self.info_color, nl=False)
        else:
            click.echo(f"  [{str(answer_number)}] ", nl=False)
        click.echo(answer)

    def prompt(self, *args, **kwargs):
        """Wrapper for click.prompt"""
        self._emit_prefix(self.warn_color, self.warning_prefix)
        return click.prompt(*args, **kwargs)

    def confirm(self, *args, **kwargs):
        """Wrapper for click.confirm"""
        self._emit_prefix(self.warn_color, self.warning_prefix)
        return click.confirm(*args, **kwargs)



from abc import abstractmethod, ABC


class AbstractSubConfig(ABC):

    @abstractmethod
    def apply_to_dict(self, dict_to_modify: dict):
        """Modify the dictionary to include the values for writing the config file."""
        pass

    @abstractmethod
    def read_from_dict(self, new_values: dict):
        """Read the values from a dictionary for reading the yaml file."""
        pass

    def __eq__(self, other):
        """Equality check by comparing the representation."""
        if not isinstance(other, self.__class__):
            return NotImplemented
        return repr(self) == repr(other)



import subprocess
import sys
from pathlib import Path

import click

from todotree.Config.Config import Config


# Addons https://click.palletsprojects.com/en/8.1.x/commands/#custom-multi-commands

class Addons:
    # Note: Does not inherit AbstractCommand!

    def __init__(self, config: Config = Config()):
        self.rv: list = []
        """List of programs."""

        self.config: Config = config
        """Configuration"""

    def run(self, invocation: tuple[str]):

        """Run the addon with the invocation. The invocation also contains any number of arguments."""
        name, *arguments = invocation
        file = Path(self.config.paths.addons_folder / name)
        if not file.exists():
            self.config.console.error(f"There is no script at {file}")
            exit(1)

        self.config.console.verbose(f"Running script at {file}")
        try:
            if name.endswith(".py"):
                result = subprocess.run([sys.executable, file, str(self.config.paths.todo_file)] + arguments, capture_output=True, text=True)
            else:
                result = subprocess.run([file, str(self.config.paths.todo_file)], capture_output=True, text=True)
        except Exception as e:
            self.config.console.error(f"Error while running the addon {name}.")
            self.config.console.error(f"The error is {e}")
            raise e

        click.echo(result.stdout, nl=False)
        if result.returncode != 0:
            click.echo(result.stderr)
        self.config.git.commit_and_push(f"addons {invocation}")

    def __call__(self, *args, **kwargs):
        raise NotImplemented # Too lazy.


import datetime
import math
import os
import pathlib
from typing import Callable

from todotree.Config.Config import Config
from todotree.Errors.ProjectFolderError import ProjectFolderError
from todotree.Errors.TodoFileNotFoundError import TodoFileNotFoundError
from todotree.Managers.AbstractManager import AbstractManager
from todotree.Task.ConsoleTask import ConsoleTask
from todotree.Task.Task import Task
from todotree.Tree import Tree


class TaskManager(AbstractManager):
    """Manages the tasks."""
    not_found_error = TodoFileNotFoundError

    @property
    def number_of_digits(self) -> int:
        """Property which defines the number of digits of the task in the task list with the highest number."""
        # The second max(, 0) is for the case that there are only project_folder tasks.
        maxi = max(max([x.i for x in self.task_list]), 0)
        return int(math.ceil(math.log(maxi + 1, 10)))

    def __init__(self, configuration: Config):
        """Initializes a task manager, which manages the tasks."""

        super().__init__()

        self.config: Config = configuration
        """Configuration settings for the application."""

        self.file = self.config.paths.todo_file
        """txt file location."""

    # Imports

    def import_tasks(self):
        """
        Import the tasks and projects from files.
        """
        super().import_tasks()
        if self.config.enable_project_folder:
            self._import_project_folder()

    def filter_t_date(self):
        """Removes Tasks from the Taskmanager where the t:date is later than today."""
        self.task_list = [t for t in self.task_list if t.t_date is None or t.t_date <= datetime.date.today()]

    def filter_block(self):
        """
        Removes Tasks from the Taskmanager where the task is blocked by some other task.
        """
        # Detect all block items.
        block_list_items = [y for x in self.task_list for y in x.blocks]
        # Blocked / Blocked by filtering
        for task in self.task_list.copy():
            # Check if there is a task which is blocked by another task.
            if any(filter(lambda x: x in block_list_items, task.blocked)):
                self.task_list.remove(task)

    def _import_project_folder(self):
        """
        Adds the directory to the project tree.
        """
        try:
            for i in pathlib.Path(self.config.paths.project_tree_folder).glob("*/"):
                proj = os.path.basename(i)
                # Check if there is a task with that project.
                if not any(proj in x.projects for x in self.task_list):
                    tmp_task = Task(-1, self.config.emptyProjectString) + f"+{proj}"
                    # Set to the highest priority, so the project bubbles up in the project tree.
                    tmp_task.priority_string = "A"
                    self.task_list.append(tmp_task)
        except FileNotFoundError as e:
            raise ProjectFolderError(f"An error occurred while processing the projects folder: {e}")

    def print_tree(self, key):
        """Prints a tree representation of the task list.
        :key: The key of the task used for the first line."""
        self.filter_t_date()
        self.filter_block()
        tasks = [ConsoleTask(task.i, task.task_string, total_digits=self.number_of_digits) for task in self.task_list]

        print_function: Callable[[str, ConsoleTask], str]
        match key:
            case "projects":
                def print_function(project_to_exclude, task):
                    return task.print_project(project_to_exclude)

                root_name = "Projects"
                empty_string = self.config.noProjectString
            case "contexts":
                def print_function(context_to_exclude, task):
                    return task.print_context(context_to_exclude)

                root_name = "Contexts"
                empty_string = self.config.noContextString
            case "due":
                def print_function(_, task):
                    return task.print_due()

                root_name = "Due Dates"
                key = "due_date_band"
                empty_string = "No due date."
            case _:
                raise NotImplementedError(key)
        tree = Tree(tasks, key, treeprint=self.config.tree_print, print_func=print_function, root_name=root_name,
                    empty_string=empty_string)
        tree.sort()
        return str(tree)

    def add_or_update_task(self, task_number, func: Callable, *args, **kwargs) -> Task:
        """
        Runs `func` on the task with `task_number`.
        :param task_number: The task number.
        :param func: The Task function.
        :param args: passed to func.
        :param kwargs: passed to func.
        :return: The updated task.
        """
        t = self.get_by_task_number(task_number)
        func(t, *args, **kwargs)
        self.write_to_file()
        return t

    def append_to_task(self, task_number: int, task_string: str) -> Task:
        """
        Appends `task_string` to the task defined by `task_number`.
        :param task_string: The string to append
        :param task_number: The number of the existing task.
        :return: The new merged task.
        """
        # Append to task.
        self.task_list[self.task_list.index(self.get_by_task_number(task_number))] += task_string
        # Write results.
        self.write_to_file()
        return self.get_by_task_number(task_number).task_string

    def filter(self, filter_string=None):
        """
        Filters the task list by whether `filter_string` occurred in the task.
        """
        self.task_list = [item for item in self.task_list if filter_string in item.task_string]
        return str(self)

    def __str__(self):
        # Prepare the lists.
        self.filter_t_date()
        self.filter_block()
        self.task_list.sort(key=lambda x: x.priority)
        # Output the list.
        return super().__str__()



from typing import Callable

from pathlib import Path

import xdg_base_dirs

from todotree.Config.Paths import Paths
from todotree.Config.ConsolePrefixes import ConsolePrefixes
from todotree.Errors.ConfigFileNotFoundError import ConfigFileNotFoundError
from todotree.Config.GitHandler import GitHandler
from todotree.Config.TreePrint import TreePrint

# Note this package is not used by write.
import yaml


class Config:
    """
    The configuration of todotree.

    For a user guide, please refer to the content in `examples/config.yaml` instead.
    Not all values are mapped 1-to-1.
    """

    def __init__(self):
        #  Main variables.
        self.config_file: Path | None = None
        """Path to the config file."""

        # Features - Enables or disables certain features.
        self.enable_project_folder: bool = True
        """A value indicating whether to enable the project folder functionality."""

        #  Localization. #
        self.wishlistName: str = "wishlist"
        self.noProjectString: str = "No Project"
        self.noContextString: str = "No Context"
        self.emptyProjectString: str = "> (A) Todo: add next todo for this."

        # Paths.
        self.paths = Paths()

        #  Status Decorators.
        self.console: ConsolePrefixes = ConsolePrefixes(True, " * ", " ! ", "!!!", "green", "yellow", "red")

        # Tree prints.
        self.tree_print: TreePrint = TreePrint(" +--", " |  ", " *--", "    ")

        # Git functionality.
        self.git = GitHandler("disabled", 1, self.paths.todo_folder, self.console)

        # Debug
        self.__yaml_object = None
        """Raw config object, for debugging."""

    def read(self, path: Path | None = None):
        """
        Loads the configuration from the given file at `path`.

        If empty, reads from default locations.
        """
        if path is not None:
            # print(f"Path is not None: {path}") # FUTURE: Verbose logging
            self.read_from_file(Path(path))
            return
        # xdg compliant directory.
        if Path(xdg_base_dirs.xdg_config_home() / "todotree" / "config.yaml").exists():
            # print(f"Path is XDG Config") # FUTURE: Verbose logging
            self.read_from_file(Path(xdg_base_dirs.xdg_config_home() / "todotree" / "config.yaml"))
            return
        # xdg compliant directory if config file is considered "data".
        if Path(xdg_base_dirs.xdg_data_home() / "todotree" / "config.yaml").exists():
            # print(f"Path is XDG DATA") # FUTURE: Verbose logging
            self.read_from_file(Path(xdg_base_dirs.xdg_data_home() / "todotree" / "config.yaml"))
            return
        # No paths: use the defaults.
        return

    def read_from_file(self, file: Path):
        """Reads and parses yaml content from `file`."""
        self.config_file = file
        try:
            self._read_from_yaml(self.config_file.read_text())
        except FileNotFoundError as e:
            raise ConfigFileNotFoundError(f"Config was not found at {self.config_file}") from e

    def _read_from_yaml(self, yaml_content: str):
        """Reads and overrides config settings defined in `yaml_content`."""
        # Convert yaml to python object.

        self.__yaml_object = yaml.safe_load(yaml_content)
        if self.__yaml_object is None:
            return
        # Map each item to the self config.
        self.__section("main", self.__section_main)
        self.__section("paths", self.__section_paths)
        self.__section("localization", self.__section_localization)
        self.__section("decorators", self.__section_decorators)
        self.__section("tree", self.__section_tree)
        self.__section("git", self.__section_git)

    def __section(self, section_name: str, section_parser: Callable):
        """
        Checks if the section exists and passes it to the sub parser.
        """
        if self.__yaml_object.get(section_name) is not None:
            self.console.verbose(f"[{section_name}] section found, reading it.")
            # Parse the actual values.
            section_parser()
        else:
            self.console.verbose(f"[{section_name}] section not found, skipping it.")

    def __section_tree(self):
        self.tree_print.read_from_dict(self.__yaml_object["tree"])

    def __section_decorators(self):
        self.console.read_from_dict(self.__yaml_object["decorators"])

    def __section_localization(self):
        self.emptyProjectString = self.__yaml_object["localization"].get("empty_project", self.emptyProjectString)
        self.wishlistName = self.__yaml_object["localization"].get("wishlist_name", self.wishlistName)
        self.noProjectString = self.__yaml_object["localization"].get("no_project", self.noProjectString)
        self.noContextString = self.__yaml_object["localization"].get("no_context", self.noContextString)

    def __section_paths(self):
        self.paths.read_from_dict(self.__yaml_object["paths"])
        self.git.todo_folder = self.paths.todo_folder

    def __section_main(self):
        self.enable_project_folder = self.__yaml_object["main"].get("enable_project_folder", self.enable_project_folder)

        # Postprocessing.
        if self.__yaml_object["main"].get("verbose", self.console.is_verbose()):
            self.console.set_verbose()
        if self.__yaml_object["main"].get("quiet", self.console.is_quiet()):
            self.console.set_quiet()

    def __section_git(self):
        self.git.read_from_dict(self.__yaml_object["git"])

    def write_config(self):
        """
        Writes the configuration file to the path specified in `self.config_file`
        """
        # Local import so that the file is not loaded in each time the program is started.
        from ruamel.yaml import YAML
        from importlib import resources as import_resources
        from .. import examples

        # https://stackoverflow.com/questions/6028000/how-to-read-a-static-file-from-inside-a-python-package
        example_config_path = import_resources.files(examples) / "config.yaml"

        with example_config_path.open("r", encoding="utf-8") as f:
            example_contents = f.read()

        yaml_structure = YAML().load(example_contents)
        # Change features.
        self.tree_print.apply_to_dict(yaml_structure["tree"])
        self.git.apply_to_dict(yaml_structure["git"])
        self.console.apply_to_dict(yaml_structure["decorators"])
        self.paths.apply_to_dict(yaml_structure["paths"])

        # Main.
        yaml_structure["main"]["verbose"] = self.console.is_verbose()
        yaml_structure["main"]["quiet"] = self.console.is_quiet()
        yaml_structure["main"]["enable_project_folder"] = self.enable_project_folder

        # Localization.
        yaml_structure["localization"]["wishlist_name"] = self.wishlistName
        yaml_structure["localization"]["no_project"] = self.noProjectString
        yaml_structure["localization"]["no_context"] = self.noContextString
        yaml_structure["localization"]["empty_project"] = self.emptyProjectString

        # Write the file.
        try:
            self.config_file.parent.mkdir(parents=True, exist_ok=True)
            self.config_file.touch()
            write_yaml_settings = YAML()
            write_yaml_settings.dump(yaml_structure, stream=self.config_file)
        except PermissionError as e:
            self.console.error("Cannot write the configuration to the following location.")
            self.console.error(self.config_file)
            self.console.error("Due to a PermissionDenied Error.")
            self.console.error("++++ Please reinstall universe and reboot ++++")
            raise e
        except Exception as e:
            self.console.error("Exception occurred while writing the configuration file.")
            self.console.error(str(e))
            raise e



import click

from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.TodoFileNotFoundError import TodoFileNotFoundError


class Filter(AbstractCommand):

    def run(self, search_term=None):
        try:
            self.config.console.info(f"Todos for term '{search_term}'")
            click.echo(self(search_term))
        except TodoFileNotFoundError as e:
            e.echo_and_exit(self.config)

    def __call__(self, search_term=None):
        self.taskManager.import_tasks()
        return self.taskManager.filter(search_term)


from pathlib import Path

import click

from todotree.Commands.AbstractCommand import AbstractCommand


class Cd(AbstractCommand):

    def run(self):
        config_path: Path = Path(self.config.paths.todo_folder)
        self.config.console.verbose("The location to the data folder is: ")

        if config_path.is_absolute():
            # Then the configured path is printed.
            click.echo(str(config_path))
            exit()
        # Then the relative path is resolved to be absolute.
        base_path: Path = Path.home()
        full_path: Path = base_path / config_path
        click.echo(str(full_path))
        exit()

    def __call__(self, *args, **kwargs):
        raise NotImplemented


from pathlib import Path

from todotree.Errors.TodotreeError import TodotreeError


class ConfigFileNotFoundError(TodotreeError):
    """
    Represents an error indicating that the config.yaml file is not found.
    """
    def echo_and_exit(self, config, config_path: Path, verbose: bool):
        """
        @param config: Config object for defaults. (This is here to prevent a circular import).
        @param config_path: The path that did not have the config.yaml
        @param verbose: Whether to add more information.
        """
        cp = config.console
        cp.warning(f"The config.yaml file could not be found at {config_path}.")
        if verbose:
            cp.warning(f"The config file should be located at {config_path}")
            cp.warning(str(self))
        cp.warning("The default options are now used.")
        exit(1)



from pathlib import Path

import click
import xdg_base_dirs

from todotree.Config.Config import Config
from todotree.Config.ConsolePrefixesInit import ConsolePrefixesInit
from todotree.Errors.GitError import GitError


class Init:
    # Note: Does not inherit AbstractCommand!

    def __init__(self):
        self.config: Config = Config()
        """Configuration."""
        self.action_queue: list[Path] = [Path("config.yaml")]
        """List of actions to execute."""
        self.git_init = False
        """Whether to do 'git init' in the folder."""
        self.homework: list[str] = []
        """List of actions the user has to do after todotree init writes the files."""
        self.console: ConsolePrefixesInit = ConsolePrefixesInit.from_console_prefixes(self.config.console)

    def run(self):
        # Check that the extras version is installed.
        try:
            import ruamel.yaml
        except ModuleNotFoundError:
            self.config.console.error("Module not found, you likely have the base version installed.")
            self.config.console.error("Install the extras version with the following command:")
            click.echo("   pip install todotree[init]")
            exit(1)
        self.intro()
        is_git_cloned = False  # self.git_clone()
        if not is_git_cloned:
            self.determine_console_prefixes()
            self.determine_config_location()
            self.determine_folder_location()
            self.enable_git()
            self.enable_project()
            self.examples()
        # Write results of the answers to the locations.
        self.console.warning("The content of config.yaml is generated.")
        self.console.warning(
            "Todotree will overwrite the following files, destroying any existing content in the files.")
        self.console.warning([self.config.paths.todo_file, self.config.paths.done_file, self.config.paths.recur_file,
                              self.config.paths.stale_file])
        self.console.warning("Todotree will add examples and defaults to the following files.")
        self.console.warning(", ".join([str(x.resolve()) for x in self.action_queue]))
        answer = self.console.confirm("Are you *sure* you want to continue?")
        if answer:
            self.config.write_config()
            self.write_example(self.config.paths.todo_file, "todo.txt")
            self.write_example(self.config.paths.done_file, "done.txt")
            self.write_example(self.config.paths.recur_file, "recur.txt")
            self.write_example(self.config.paths.stale_file, "stale.txt")
        else:
            self.console.error("Aborted the initialization, please run `todotree init` again.")
            exit()
        # Homework section.
        if self.git_init:
            self.run_git_init()
        self.console.info("Written the files.")
        if self.homework:
            self.console.info("The following things need to be configured yet.")
        for task in self.homework:
            self.console.info(task)

    def intro(self):
        self.console.info("This will configure todotree and generate the necessary files to run it.")
        self.console.info("If at any point you wish to stop, you can do so with Ctrl-C.")
        self.console.info("Also, all options can be changed afterwards in the config.yaml if you change your mind.")

    def determine_config_location(self):
        question = "Where do you want to store config.yaml?"
        answers = [
            Path(xdg_base_dirs.xdg_config_home() / "todotree" / "config.yaml"),
            Path(xdg_base_dirs.xdg_data_home() / "todotree" / "config.yaml"),
        ]
        answer = self.console.prompt_menu(question, answers,
                                          "Custom location (Note: you'll need to supply it each time "
                                          "with the --config-file option).")
        if isinstance(answer, int):
            self.config.config_file = answers[answer]
        else:
            self.config.config_file = Path(answer)
            self.console.warning(
                "Note: You need to supply todotree each time with the config location using --config-file"
            )
            self.console.warning("It is advised to alias this in your profile for example in ~/.profile or $PROFILE.")
            self.homework.append("Custom config file: Configure $PROFILE or ~/.profile have --config-file.")
        self.console.info(f"Set the config location to {self.config.config_file}")

    def determine_console_prefixes(self):
        self.console.enable_colors = self.console.confirm(text="Do you want colors on your console?")
        self.config.console.enable_colors = self.console.enable_colors
        question = "How do you want the decorations?"
        answer: int = -1
        if self.console.enable_colors:
            confirmed = False
            while not confirmed:
                answer, confirmed = self.__prompt_prefix_colors(question)
        else:
            answers = [
                (f"Default      | {self.console.info_prefix} | {self.console.warning_prefix} |"
                 f" {self.console.error_prefix} |"),
                "Gentoo Style |  *  |  *  |  *  |",
            ]
            answer = int(self.console.prompt_menu(question, answers))
        if answer == 1:
            # Then we set gentoo style in both current shell and config.
            (self.console.info_prefix, self.console.warning_prefix, self.console.error_prefix,
             self.config.console.info_prefix, self.config.console.warning_prefix, self.config.console.error_prefix) \
                = [" * "] * 6

    def __prompt_prefix_colors(self, question):
        """Displays the options in color as well."""
        self.console.warning(question)

        def first():
            click.secho("[0]", fg=self.console.info_color, nl=False)
            click.echo(" Default      | ", nl=False)
            click.secho(self.console.info_prefix, fg=self.console.info_color, nl=False)
            click.echo(" | ", nl=False)
            click.secho(self.console.warning_prefix, fg=self.console.warn_color, nl=False)
            click.echo(" | ", nl=False)
            click.secho(self.console.error_prefix, fg=self.console.error_color)

        def second():
            click.secho("[1]", fg=self.console.info_color, nl=False)
            click.echo(" Gentoo Style | ", nl=False)
            click.secho(" * ", fg=self.console.info_color, nl=False)
            click.echo(" | ", nl=False)
            click.secho(" * ", fg=self.console.warn_color, nl=False)
            click.echo(" | ", nl=False)
            click.secho(" * ", fg=self.console.error_color)

        first()
        second()
        # Ask the question.
        answer = int(self.console.prompt(text="Your Answer", type=click.Choice(["0", "1"])))
        first() if answer == 0 else second()
        confirmed = self.console.confirm(text=f"Is above correct?")
        return answer, confirmed

    def determine_folder_location(self):
        question = "Where do you want to store the other files?"
        answers = [
            Path(xdg_base_dirs.xdg_config_home() / "todotree"),
            Path(xdg_base_dirs.xdg_data_home() / "todotree"),
        ]
        answer = self.console.prompt_menu(question, answers, custom_answer="Custom folder location.")
        if isinstance(answer, int):
            self.config.paths.todo_folder = answers[answer]
        else:
            self.config.paths.todo_folder = Path(answer).resolve()
        self.console.info(f"Set the folder location to {self.config.paths.todo_folder}")

    def enable_git(self):
        enable_git = self.console.confirm(text="Do you want to enable git?")
        if not enable_git:
            git_mode = "disabled"
            self.console.warning("Note, if you want to configure it in the future, "
                                 "you need to run `git init` in the folder.")
        else:
            # Add git to the list of things to do
            self.git_init = True
            enable_remote = self.console.confirm(text="Do you want to work with a remote repo as well?")
            git_mode = "full" if enable_remote else "local"
            if git_mode == "full":
                self.config.console.warning("You will need to configure the remote repo later.")
                self.homework.append("Configure remote git repo.")
        try:
            self.config.git.git_mode = git_mode
        except GitError as e:
            self.console.error("Setting git raised an error, you'll need to fix this later. The error is:")
            self.console.error(str(e))
            self.homework.append("Configure git itself.")


    def examples(self):
        """Ask about the example files."""
        questions = [
            ("Do you want to have your todo.txt file filled with some examples?", self.config.paths.todo_file),
            ("Do you want to have your stale.txt file filled with some examples?", self.config.paths.stale_file),
            ("Do you want to have your recur.txt file filled with some examples?", self.config.paths.recur_file),
        ]
        for question, file in questions:
            if self.console.confirm(question):
                self.action_queue.append(file)

    def enable_project(self):
        self.console.info("The project_directory functionality adds the projects in a given folder.")
        self.console.info("It will also add tasks if there is no task with the given project name in todo.txt,")
        self.console.info("reminding you that that project is stalled.")
        self.config.enable_project_folder = self.console.confirm(
            "Do you want to enable the project_directory functionality?"
        )
        if self.config.enable_project_folder:
            confirmed = False
            while not confirmed:
                if self.config.enable_project_folder:
                    answer = self.console.prompt(
                        "What will the location of your projects be?", default=self.config.paths.project_tree_folder
                    )
                    self.config.project_tree_folder = Path(answer).resolve()
                if not Path(self.config.paths.todo_folder).exists():
                    self.console.warning("Project folder does not exist yet.")
                confirmed = self.console.confirm("Is the project directory correct?")
            if not Path(self.config.paths.todo_folder).exists():
                self.homework.append(f"Create a project folder at {self.config.project_tree_folder} and add projects.")

    def git_clone(self):
        answer = self.console.confirm("Do you already have an existing todotree folder on git?")
        if answer:
            self.console.prompt("Please enter the git clone url to clone from.")
            # FUTURE: implement git clone.
        return answer

    def write_example_done(self):
        """Create a done.txt file in the configured location."""
        self.config.paths.done_file.touch()

    def write_example(self, file: Path, example_file: str):
        """
        Create a file with or without contents.
        :file: destination file path of the content.
        :example_file: name of the file in the example folder in the source code.
        """
        if file not in self.action_queue:
            # Do not write content.
            example_contents = ""
        else:
            # Import the example content.
            from importlib import resources as import_resources
            from .. import examples
            example_todo_path = import_resources.files(examples) / example_file
            example_contents = example_todo_path.read_text()
        file.parent.mkdir(parents=True, exist_ok=True)
        file.write_text(example_contents)

    def run_git_init(self):
        """Runs git init in the configured folder."""
        self.config.git.git_init()



# NOTE: this file cannot be in a class. See: https://github.com/pallets/click/issues/601
# https://click.palletsprojects.com/en/8.1.x/commands/#nested-handling-and-contexts
import click

# Commands are imported when the function is run, to increase performance.
from todotree.Managers.TaskManager import TaskManager
from todotree.Config.Config import Config
from todotree.MainUtils import MainUtils


@click.group()
@MainUtils.common_options
def root(ctx: click.Context, **kwargs):
    """
    The main list of todotree's command.

    For more information on a command, use `todotree COMMAND --help`.
    """
    # ^ This text also shows up in the help command.
    # Parse options given before the command.
    # So the `--todo-file task.txt` option in the help text example.
    ctx.ensure_object(dict)
    ctx.obj["config"] = Config()
    MainUtils.parse_common_options(ctx.obj["config"], **kwargs)


@root.command("add", short_help="Add a task to the task list")
@click.argument("task", type=str, nargs=-1)
@MainUtils.common_options
def add(ctx: click.Context, task: tuple, **kwargs):
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Add import Add

    Add(ctx.obj["config"], TaskManager(ctx.obj["config"])).run(task)


@root.command("addx", short_help="Add a task and immediately mark it as done")
@click.argument("task", type=str, nargs=-1)
@MainUtils.common_options
def add_x(ctx: click.Context, task: tuple, **kwargs):
    """
    Adds a completed task to done.txt. The task is not added to todo.txt.
    :param task: The task to add to done.txt.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.AddX import AddX

    AddX(ctx.obj["config"], TaskManager(ctx.obj["config"])).run(task)


@root.command("append", short_help="Append `append_string` to `task_nr`")
@click.argument("task_nr", type=int)
@click.argument("append_string", type=str, nargs=-1)
@MainUtils.common_options
def append(ctx: click.Context, task_nr: int, append_string: tuple[str], **kwargs):
    """
    Appends the contents of append_string to the task represented by task_nr.
    A space is inserted between the two tasks, so you do not have to worry that words aretogether.

    Example: todotree append 1 "some text"
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Append import Append

    Append(ctx.obj["config"], TaskManager(ctx.obj["config"])).run(task_nr, append_string)


@root.command("cd", short_help="Print directory of the todo.txt directory")
@MainUtils.common_options
def cd(ctx: click.Context, **kwargs):
    """print directory of the todo.txt directory"""
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Cd import Cd

    Cd(ctx.obj["config"], TaskManager(ctx.obj["config"])).run()


@root.command("context", short_help="List task in a tree, by context")
@MainUtils.common_options
def context(ctx: click.Context, **kwargs):
    """list a tree, of which the first node is the context, the second nodes are the tasks"""
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Context import Context

    Context(ctx.obj["config"], TaskManager(ctx.obj["config"])).run()


@root.command("do", short_help="Mark task as done and move it to the done.txt")
@click.argument("task_numbers", type=list, nargs=-1)  # type=list[int]
@MainUtils.common_options
def do(ctx: click.Context, task_numbers: list[tuple], **kwargs):
    """
    Mark tasks as done, therefor moving them to done.txt with a date stamp of today.
    :param task_numbers: The list of tasks which are completed.
    """

    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Do import Do

    Do(ctx.obj["config"], TaskManager(ctx.obj["config"])).run(task_numbers)


@root.command("due", short_help="List tasks by their due date")
@MainUtils.common_options
def due(ctx: click.Context, **kwargs):
    """List tasks in a tree by their due date."""
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Due import Due

    Due(ctx.obj["config"], TaskManager(ctx.obj["config"])).run()


@root.command("edit", short_help="Open the todo.txt in an editor.")
@click.argument("file", default="todo",
                type=click.Choice(["todo", "done", "stale", "recur"]))
@MainUtils.common_options
def edit(ctx: click.Context, **kwargs):
    """
    Open the file in an editor for manual editing of tasks.

    This is useful when you need to modify a lot of tasks, which would be complicated when doing it with todotree.
    The second argument defines which file you want to edit, defaults to todo.txt.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Edit import Edit

    Edit(ctx.obj["config"], TaskManager(ctx.obj["config"])).run(kwargs["file"])


@root.command("filter", short_help="Only show tasks containing the search term.")
@click.argument("search_term")
@MainUtils.common_options
def filter_list(ctx: click.Context, search_term, **kwargs):
    """
    Only show tasks which have search term in them. This can also be a keyword.

    :param search_term: The term to search.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Filter import Filter

    Filter(ctx.obj["config"], TaskManager(ctx.obj["config"])).run(search_term)


@root.command("help", short_help="Print help text.")
@click.pass_context
def help_text(ctx: click.Context):
    print(ctx.parent.get_help())


@root.command("init", short_help="Initialize folder for first use.")
@MainUtils.common_options
def init(ctx: click.Context, **kwargs):
    """
    Initializes the files and folders needed for a functioning todotree according to the user's prompts.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Init import Init

    Init().run()


@root.command("list", short_help="List tasks")
@MainUtils.common_options
def list_tasks(ctx: click.Context, **kwargs):
    """
    Print a flat list of tasks, sorted by their priority.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.List import List

    MainUtils.parse_common_options(ctx.obj["config"], **kwargs)
    List(ctx.obj["config"], TaskManager(ctx.obj["config"])).run()


@root.command("list_done", short_help="List tasks which are marked as done")
@MainUtils.common_options
def list_done(ctx: click.Context, **kwargs):
    """
    List tasks which are marked as done. The numbers can be used with the revive command.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.ListDone import ListDone

    ListDone(ctx.obj["config"], TaskManager(ctx.obj["config"])).run()


@root.command("print_raw", short_help="Print todo.txt without any formatting or filtering")
@MainUtils.common_options
def print_raw(ctx: click.Context, **kwargs):
    """
    Output the todo.txt without any processing.
    This is equivalent to `cat $(todo cd)/todo.txt` in bash.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.PrintRaw import PrintRaw

    PrintRaw(ctx.obj["config"], TaskManager(ctx.obj["config"])).run()


@root.command("priority", short_help="Set new priority to task")
@click.argument("task_number", type=int)
@click.argument("new_priority", type=str)
@MainUtils.common_options
def priority(ctx: click.Context, task_number, new_priority, **kwargs):
    """
    Adds or updates the priority of the task.
    :param task_number: The task to re-prioritize.
    :param new_priority: The new priority.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Priority import Priority

    Priority(ctx.obj["config"], TaskManager(ctx.obj["config"])).run(task_number, new_priority)


@root.command("project", short_help="Print tree by project")
@MainUtils.common_options
def project(ctx: click.Context, **kwargs):
    """
    Print the task list in a tree by project.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Project import Project

    Project(ctx.obj["config"], TaskManager(ctx.obj["config"])).run()


@root.command("recur", short_help="Add recurring tasks defined in recur.txt")
@click.option("--date", help="Use this date instead of the one in the timestamp file.")
@MainUtils.common_options
def recur(ctx: click.Context, date: str | None, **kwargs):
    """
    Add tasks according to recur.txt.
    --date: Use this date instead of the one in the timestamp file.

    In recur.txt you can add tasks that have to be done periodically, such as doing taxes.
    The format in recur.txt is:

        `<start date> ; interval ; <end date> ; <task string>`

    The end date is optional. This is a valid line for example:
    2020-01-01 ; yearly ; (C) Celebrate the new year.

    This is certainly not the only program which can do these.
    Different solutions to this command exist, here are a few alternatives:
     - https://www.rememberthemilk.com/
     - https://github.com/smartlitchi/ice-recur
    You can use those with the `addons` command after installing a suitable addon script.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Recur import Recur
    Recur(ctx.obj["config"], TaskManager(ctx.obj["config"])).run(date=date)


@root.command("revive", short_help="Revive a task that was accidentally marked as done.")
@click.argument("done_number", type=int)
@MainUtils.common_options
def revive(ctx: click.Context, done_number, **kwargs):
    """
    Move a task from done.txt to todo.txt.
    The `done_number` can be found using the `list_done` command.
    :param done_number: The number of the task to revive.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Revive import Revive

    Revive(ctx.obj["config"], TaskManager(ctx.obj["config"])).run(done_number)


@root.command("schedule", short_help="Hide task until date.")
@click.argument("task_number", type=int)
@click.argument("new_date", type=str, nargs=-1)
@MainUtils.common_options
def schedule(ctx: click.Context, task_number: int, new_date: tuple[str], **kwargs):
    """
    hide the task until the date given. If new_date is not in ISO format (yyyy-mm-dd) such as "Next Wednesday",
    then it tries to figure out the date with the `date` program, which is only in linux.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Schedule import Schedule

    Schedule(ctx.obj["config"], TaskManager(ctx.obj["config"])).run(new_date, task_number)


@root.group("stale", short_help="Manage the stale.txt task list.")
@MainUtils.common_options
def stale(ctx: click.Context, **kwargs):
    # process for stale.txt
    pass


@stale.command("list", short_help="List the stale tasks.")
@MainUtils.common_options
def stale_list(ctx: click.Context, **kwargs):
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Stale_list import StaleList
    StaleList(ctx.obj["config"], TaskManager(ctx.obj["config"])).run()


@stale.command("revive", short_help="Move task from stale.txt to todo.txt")
@click.argument("task_number", type=int)
@MainUtils.common_options
def stale_revive(ctx: click.Context, task_number: int, **kwargs):
    """Remove task with index `task_number` from stale.txt and adds it back to todo.txt."""
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Stale_revive import StaleRevive
    StaleRevive(ctx.obj["config"], TaskManager(ctx.obj["config"])).run(task_number)


@stale.command("add", short_help="Move task from todo.txt to stale.txt")
@click.argument("task_number", type=int)
@MainUtils.common_options
def stale_add(ctx: click.Context, task_number: int, **kwargs):
    """Removes the task with the index `task_number` from todo.txt and adds it to stale.txt."""
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Stale_add import StaleAdd
    StaleAdd(ctx.obj["config"], TaskManager(ctx.obj["config"])).run(task_number)


@root.command("version", short_help="Show the version of TodoTree.")
@MainUtils.common_options
def version(ctx: click.Context, **kwargs):
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Version import Version

    Version(ctx.obj["config"], TaskManager(ctx.obj["config"])).run()


@root.command("addons", short_help="Run an addon script", context_settings={"ignore_unknown_options": True})
@click.argument("command", nargs=-1, type=str)
@MainUtils.common_options
def addons_command(ctx: click.Context, command: str, **kwargs):
    """
    Run an addon script.
    The script must be in the addons_folder. It can be any language you like: It does not have to be python.
    However, it must have the executable bit set.

    :param command: The script/command to run.
    """
    MainUtils.initialize(ctx.obj["config"], **kwargs)
    from todotree.Commands.Addons import Addons

    Addons(ctx.obj["config"]).run(command)


#  End Region Command Definitions.
#  Setup Click

CONTEXT_SETTINGS: dict = dict(help_option_names=["-h", "--help"])
"""Click context settings. See https://click.palletsprojects.com/en/8.1.x/complex/ for more information."""
cli: click.CommandCollection = click.CommandCollection(sources=[root], context_settings=CONTEXT_SETTINGS)
"""Command Collection defining defaults. https://click.palletsprojects.com/en/8.1.x/api/#click.CommandCollection ."""

if __name__ == "__main__":
    cli(obj={})



from todotree.Task.DoneTask import DoneTask
from todotree.Task.Task import Task


class ConsoleTask(Task):
    """
    Task class for printing to the console.
    """

    @staticmethod
    def from_done_task(task: DoneTask, total_digits) -> 'ConsoleTask':
        ct = ConsoleTask(task.i, task.task_string, total_digits)
        ct.date_marked_done = task.date_marked_done
        return ct

    def __init__(self, i: int, task_string: str, total_digits=1):
        super().__init__(i, task_string)
        self.total_digits = total_digits
        """Represents the total number of digits that `i` must have."""

        self.parts: list = []
        """The parts of the final string for printing."""

    def __str__(self):
        """Base representation."""
        self.__add_i()
        self.__add_date_marked_done()
        self.__add_priority()
        self.__add_other()
        self.__add_due_date()
        self.__add_projects()
        self.__add_context()
        self.__add_blocks()
        return " ".join(self.parts)

    def print_due(self):
        """
        Print the task for task manager due dates.
        :return: the string suitable for printing the due date tree.
        """
        self.__add_i()
        self.__add_priority()
        self.__add_other()
        self.__add_projects()
        self.__add_context()
        self.__add_blocks()
        self.__add_due_date()
        return " ".join(self.parts)

    def print_project(self, excluded_project=None):
        """
        Strips the string for task_manager.project_tree
        :return: a string useful for printing the project tree.
        """
        self.__add_i()
        self.__add_priority()
        self.__add_other()
        self.__add_due_date()
        self.__add_projects(excluded_project)
        self.__add_context()
        self.__add_blocks()
        return " ".join(self.parts)

    def print_context(self, excluded_context=None):
        """
        Strips the string for task_manager.project_tree
        :return: a string useful for printing the project tree.
        """
        self.__add_i()
        self.__add_priority()
        self.__add_other()
        self.__add_due_date()
        self.__add_projects()
        self.__add_context(excluded_context)
        self.__add_blocks()
        return " ".join(self.parts)

    def __add_i(self):
        """Add the number to `parts`."""
        if self.i > 0:
            self.parts.extend([str(self.i).zfill(self.total_digits)])

    def __add_context(self, excluded_context=None):
        """
        Add the contexts to `parts`.
        :param excluded_context: The context to exclude.
        """
        self.parts.extend([f"@{c}" for c in self.contexts if c != excluded_context])

    def __add_projects(self, excluded_project=None):
        """
        Add the projects to `parts`.
        :param excluded_project: The project to exclude.
        """
        self.parts.extend([f"+{p}" for p in self.projects if p != excluded_project])

    def __add_other(self):
        """Add the non-parsed material to `parts`"""
        if self.other_string:
            self.parts.extend([self.other_string])

    def __add_priority(self):
        """Add the formatted priority to `parts`."""
        if self.priority_string:
            self.parts.extend([f"({self.priority_string})"])

    def __add_blocks(self):
        """Add the blocks to `parts`."""
        self.parts.extend([f"bl:{bl}" for bl in self.blocks])

    def __add_due_date(self):
        """Add the due date to `parts`."""
        if self.due_date is not None:
            self.parts.extend(["due:" + self.due_date.strftime("%Y-%m-%d")])

    def __add_date_marked_done(self):
        if self.date_marked_done is not None:
            self.parts.extend(["x " + self.date_marked_done])



import click

from todotree.Config.AbstractSubConfig import AbstractSubConfig


class ConsolePrefixes(AbstractSubConfig):
    """Class for handling the prefixes in the terminal.

    The class is somewhat similar to a simple logger. It prints messages with a configured prefix.
    """

    def __init__(self, enable_colors: bool,
                 info_prefix: str, warning_prefix: str, error_prefix: str,
                 info_color: str, warn_color: str, error_color: str):
        """
        Initialize a new Console Prefix.
        :param enable_colors: Whether to enable colors.
        :param info_prefix: The prefix for information messages.
        :param warning_prefix: The prefix for warning messages.
        :param error_prefix: The prefix for error messages.
        """
        self.enable_colors = enable_colors
        self.error_prefix = error_prefix
        self.warning_prefix = warning_prefix
        self.info_prefix = info_prefix
        self.info_color = info_color
        self.warn_color = warn_color
        self.error_color = error_color

        self.__detail = 0
        """An integer indicating how much detail is printed."""

    def verbose(self, message):
        """
        Print an info message if verbose is turned on.
        """
        if self.__detail > 0:
            self.info(message)

    def info(self, message):
        """
        Print an information message to console if quiet is not on.
        :param message: The message to print.
        """
        if self.__detail >= 0:
            self.__emit(self.info_prefix, self.info_color, message)

    def warning(self, message: str):
        """
        Print a warning message to the console.
        :param message: The message to print.
        """
        self.__emit(self.warning_prefix, self.warn_color, message)

    def error(self, message: str):
        """
        Print an error message to the console.
        :param message: The message to print.
        """
        self.__emit(self.error_prefix, self.error_color, message)

    def __emit(self, prefix: str, color: str, message: str):
        """
        Print a message with a prefix.
        :param prefix: The prefix of the message.
        :param message: The message itself.
        """
        # Emit the prefix.
        self._emit_prefix(color, prefix)
        # Emit the message.
        click.echo(message)

    def _emit_prefix(self, color, prefix):
        """Print the prefix with color if enabled."""
        if self.enable_colors:
            click.secho(prefix, fg=color, nl=False)
        else:
            click.echo(prefix, nl=False)

    def is_verbose(self):
        return self.__detail > 0

    def set_verbose(self):
        """Turns on verbose messaging."""
        self.__detail = 1

    def set_quiet(self):
        """Disables info messages."""
        self.__detail = -1

    def is_quiet(self):
        """Return whether quiet is turned on."""
        return self.__detail < 0

    def read_from_dict(self, new_values: dict):
        self.enable_colors = new_values.get('enable_colors', self.enable_colors)
        self.info_prefix = new_values.get('info', self.info_prefix)
        self.warning_prefix = new_values.get('warning', self.warning_prefix)
        self.error_prefix = new_values.get('error', self.error_prefix)
        self.info_color = new_values.get('info_color', self.info_color)
        self.warn_color = new_values.get('warning_color', self.warn_color)
        self.error_color = new_values.get('error_color', self.error_color)

    def apply_to_dict(self, dict_to_modify: dict):
        dict_to_modify['enable_colors'] = self.enable_colors
        dict_to_modify['info'] = self.info_prefix
        dict_to_modify['warning'] = self.warning_prefix
        dict_to_modify['error'] = self.error_prefix
        dict_to_modify['info_color'] = self.info_color
        dict_to_modify['warning_color'] = self.warn_color
        dict_to_modify['error_color'] = self.error_color

    def __repr__(self):
        return f"ConsolePrefixes({', '.join([str(x) for x in list(self.__dict__.values())])})"



import datetime
import re
from itertools import chain
from pathlib import Path

from todotree.Config.Config import Config
from todotree.Errors.RecurDateStampError import RecurDateStampError
from todotree.Errors.RecurParseError import RecurParseError
from todotree.Task.RecurTask import RecurTask
from todotree.Managers.TaskManager import TaskManager


class RecurManager:
    """Manages the recur.txt tasks."""

    comment_regex = re.compile(r"^\s*#")
    """Regular expression to detect comments."""

    recur_timestamp_filename = ".recur.timestamp"
    """Name of the timestamp file."""

    def __init__(self, config: Config):
        self.config: Config = config
        self.task_list: list[RecurTask] = []
        self.task_manager: TaskManager = TaskManager(self.config)
        self.last_date: datetime.date | None = None
        self.timestamp_file: Path = self.config.paths.todo_folder / self.recur_timestamp_filename
        self.is_manual = False
        # If supplied with the

    def import_tasks(self):
        """Import the tasks from recur.txt."""
        error = False
        with self.config.paths.recur_file.open("r") as f:
            for line_number, recur_task in enumerate(f.readlines()):
                if not recur_task.strip():
                    # Empty line, continue.
                    continue
                if self.comment_regex.match(recur_task):
                    # Comment, continue.
                    continue
                try:
                    task = RecurTask(recur_task)
                    self.task_list.append(task)
                except RecurParseError as e:
                    self.config.console.error(f"Error parsing Recur on line {line_number}.")
                    self.config.console.error(str(e))
                    error = True
                    continue
                if task.end_date < task.start_date:
                    self.config.console.warning(f"Recur task on line {line_number} is ending before it is starting.")
                    self.config.console.warning("This task will never get added.")
        if error:
            raise RecurParseError("One or more tasks were invalid, aborted adding recur tasks.")
        # Also import the tasks itself to compare later.
        self.task_manager.import_tasks()

    def add_to_todo(self):
        """Add the imported tasks to todo.txt"""
        # Check time of last time recur ran.
        self._check_last_time_run()
        # If already run, just exit.
        if self.last_date == datetime.date.today():
            self.config.console.info("Recur has already run today.")
            return

        # Filter the tasks that need to be added.
        self.filter_by_date()
        self.filter_by_existing()
        # Add the remaining Recur Tasks to the task list.
        added_tasks = self.task_manager.add_tasks_to_file([x.to_task(i) for i, x in enumerate(self.task_list, start=1)])
        if added_tasks:
            self.config.console.info("Added the following tasks:")
        else:
            self.config.console.info("No recurring task to add.")
        for i, task in enumerate(added_tasks):
            self.config.console.info(f"{task}")

    def filter_by_date(self):
        """Removes tasks that are not allowed to be added because of their date information."""
        new_list = []
        for task in self.task_list:
            if task.start_date > datetime.date.today():
                # Recur task has not started yet.
                continue
            if task.end_date < datetime.date.today():
                # Recur task has ended.
                continue
            match task.interval:
                case RecurTask.Recurrence.Never:
                    if self.last_date < task.start_date < datetime.date.today():
                        new_list.append(task)
                    elif self.last_date < task.end_date < datetime.date.today():
                        new_list.append(task)
                case RecurTask.Recurrence.Daily:
                    # Then it should always be added.
                    new_list.append(task)
                case RecurTask.Recurrence.Weekly:
                    self.__match_week(new_list, task, datetime.date.today())
                case RecurTask.Recurrence.Monthly:
                    self.__match_month(new_list, task, datetime.date.today())
                case RecurTask.Recurrence.Yearly:
                    self.__match_year(new_list, task, datetime.date.today())
        # End loop
        self.task_list = new_list

    def __match_week(self, new_list, task, today):
        # Count up until the week is further than last date.
        date = task.start_date
        while date < self.last_date:
            date = date + datetime.timedelta(days=7)
        if date < today:
            new_list.append(task)

    def __match_month(self, new_list, task, today):
        if self.last_date.day < today.day:
            # case: 12 13 14 15
            if task.start_date.day in range(self.last_date.day, today.day):
                new_list.append(task)
        else:
            # case: 30 31 1 2 3 4 5
            if task.start_date.day in chain(range(self.last_date.day, 31), range(1, today.day)):
                new_list.append(task)

    def __match_year(self, new_list, task, today):
        if self.last_date.year < today.year:
            # case: 2022 2023
            if task.start_date.month in range(self.last_date.month):
                # Check in last year
                if self.last_date < datetime.date(self.last_date.year, task.start_date.month,
                                                  task.start_date.day) < today:
                    new_list.append(task)
            else:
                # Check in new year.
                if self.last_date < datetime.date(today.year, task.start_date.month, task.start_date.day) < today:
                    new_list.append(task)
        else:
            # Both in same year.
            if self.last_date < datetime.date(self.last_date.year, task.start_date.month, task.start_date.day) < today:
                new_list.append(task)

    def filter_by_existing(self):
        """Removes tasks that already exist in todo.txt"""
        existing_tasks = [x.task_string for x in self.task_manager.task_list]
        self.task_list = [item for item in self.task_list if item.task_string not in existing_tasks]

    def _check_last_time_run(self):
        if self.last_date is not None:
            # Then it was already checked.
            return
        if not self.timestamp_file.exists():
            # Then run from the beginning of time.
            self.last_date = datetime.date(1, 1, 1)
            return
        timestamp = self.timestamp_file.read_text()
        try:
            self.last_date = datetime.date.fromisoformat(timestamp.strip())
        except ValueError as e:
            RecurDateStampError().echo_and_exit(self.config, timestamp)
            exit(1)
        if self.config.console.is_verbose():
            self.config.console.verbose(f"Last run was at date {self.last_date}")

    def _set_last_time_run(self):
        if not self.timestamp_file.exists():
            self.timestamp_file.touch()
        # This changes the mtime.
        self.timestamp_file.write_text(datetime.date.today().isoformat())



import datetime

from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.RecurParseError import RecurParseError
from todotree.Managers.RecurManager import RecurManager


class Recur(AbstractCommand):

    def run(self, date: str | None = None):
        if not self.config.paths.recur_file.exists():
            self.config.console.error("recur.txt not found, nothing to add.")
            self.config.console.error("It should be at the following location:")
            self.config.console.error(self.config.paths.recur_file)
            exit(1)
        if date is not None:
            # Parse the date into a proper date.
            try:
                date_proper = datetime.date.fromisoformat(date.strip())
            except ValueError:
                self.config.console.error(f"date {date} is not a valid iso date.")
                exit(1)
        else:
            date_proper = None

        try:
            self(date_proper)
        except RecurParseError as e:
            self.config.console.error(str(e))
            exit(1)
        self.config.git.commit_and_push("recur")

    def __call__(self, date: datetime.date | None = None, *args, **kwargs):
        rm = RecurManager(self.config)
        if date is not None:
            # This will override the timestamp file checking.
            rm.last_date = date
        rm.import_tasks()
        rm.add_to_todo()
        rm._set_last_time_run()



import re
import subprocess

from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.TodoFileNotFoundError import TodoFileNotFoundError
from todotree.Task.Task import Task


class Schedule(AbstractCommand):
    def run(self, new_date=None, task_number=None):
        # Disable fancy imports, because they do not have t dates.
        self.config.enable_project_folder = False
        # Convert
        date = " ".join(new_date)
        date_pattern = re.compile(r"(\d{4})-(\d{2})-(\d{2})")
        if not date_pattern.match(date):
            self.config.console.verbose(f"Attempt to parse {date} with the `date` program.")
            # Try to use the `date` utility.
            dat = subprocess.run(
                ["date", "-d " + date, "+%F "],
                capture_output=True,
                encoding="utf-8"
            )
            if dat.returncode > 0:
                self.config.console.error(f"The date {new_date} could not be parsed.")
                exit(1)
            date = dat.stdout.strip()
        try:
            self.taskManager.import_tasks()
            self.config.console.info(f"Task {task_number} hidden until {date}")
            updated_task = self.taskManager.add_or_update_task(task_number, Task.add_or_update_t_date, date)
            self.config.console.info(str(updated_task))
        except TodoFileNotFoundError as e:
            e.echo_and_exit(self.config)
        self.config.git.commit_and_push("schedule")

    def __call__(self, *args, **kwargs):
        NotImplemented


"""
Todotree development documentation.


"""
# Note: above text is also the welcome screen of the pdoc generated documentation.

# Expose modules/functions of the module.
from todotree.Managers.TaskManager import TaskManager
from todotree.Config.Config import Config

# Expose commands of the module.
from todotree.Commands import *


import abc
from abc import ABC

from todotree.Config.Config import Config
from todotree.Managers.TaskManager import TaskManager


class AbstractCommand(ABC):
    """Abstract Command class to be used as a base class for all command."""

    def __init__(self, config: Config, task_manager: TaskManager):
        """
        Initializes a new AbstractCommand.
        :param config: The configuration of the application.
        :param task_manager: The task manager.
        """
        self.taskManager = task_manager
        self.config = config

    @abc.abstractmethod
    def run(self) -> None:
        """Runs the command. Results are emitted through click.echo."""
        pass

    @abc.abstractmethod
    def __call__(self):
        """
        Runs the command and returns the result.
        """
        pass



from todotree.Config.Config import Config

from todotree.Errors.DoneFileNotFoundError import DoneFileNotFoundError
from todotree.Managers.AbstractManager import AbstractManager
from todotree.Task.ConsoleTask import ConsoleTask
from todotree.Task.DoneTask import DoneTask


class DoneManager(AbstractManager):
    """Manages the tasks which are done."""

    not_found_error = DoneFileNotFoundError

    def __init__(self, configuration: Config):
        super().__init__()

        self.task_list: list[DoneTask] = []
        """List of the tasks."""

        self.config: Config = configuration
        """Configuration settings for the application."""

        self.file = self.config.paths.done_file
        """txt file location."""

    def import_tasks(self):
        """Imports the tasks from the database file."""
        try:
            with self.file.open('r') as f:
                content = f.readlines()
                for i, task in enumerate(content):
                    # Skip empty lines.
                    if task.strip() == "":
                        continue
                    # The only line that is different compared to the abstract one.
                    self.task_list.append(DoneTask(i + 1, task.strip()))
        except FileNotFoundError as e:
            raise self.not_found_error("") from e

    def __str__(self):
        self.task_list.reverse()
        s = ""
        for tsk in [ConsoleTask.from_done_task(task, self.number_of_digits) for task in self.task_list]:
            s += str(tsk) + "\n"
        return s



import click

from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.TodoFileNotFoundError import TodoFileNotFoundError


class List(AbstractCommand):
    def run(self):
        try:
            message = self()
        except TodoFileNotFoundError as e:
            e.echo_and_exit(self.config)
        self.config.console.info("Todos")
        click.echo(message)

    def __call__(self, *args, **kwargs):
        self.taskManager.import_tasks()
        return self.taskManager


from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.DoneFileNotFoundError import DoneFileNotFoundError
from todotree.Errors.NoSuchTaskError import NoSuchTaskError
from todotree.Errors.TodoFileNotFoundError import TodoFileNotFoundError
from todotree.Managers.DoneManager import DoneManager
from todotree.Task.DoneTask import DoneTask


class Do(AbstractCommand):

    def __call__(self, task_numbers: list[int]) -> list[DoneTask]:
        # Marking something as Done cannot be done with fancy imports
        # So we disable them.
        self.config.enable_project_folder = False
        self.taskManager.import_tasks()
        completed_tasks = self.taskManager.remove_tasks_from_file(task_numbers)
        done_manager = DoneManager(self.config)
        done_manager.import_tasks()
        done_tasks = [DoneTask.task_to_done(task) for task in completed_tasks]
        done_manager.add_tasks_to_file(done_tasks)
        return done_tasks

    def run(self, task_numbers: list[tuple]):
        # Convert to ints. Task numbers is a list of tuples. Each tuple contains one digit of the number.
        new_numbers: list[int] = []
        for task_tuple in task_numbers:
            new_number: str = ""
            for task_digit in task_tuple:
                new_number += task_digit
            new_numbers.append(int(new_number))
        try:
            done_tasks = self(new_numbers)
        except TodoFileNotFoundError as e:
            e.echo_and_exit(self.config)
            exit(1)  # IDE hinting.
        except NoSuchTaskError as e:
            self.config.console.error(e.message)
            exit(1) # IDE hinting.
        except DoneFileNotFoundError as e:
            e.echo_and_exit(self.config)
            exit(1)  # IDE hinting.
        # Print the results
        self.config.console.info("Tasks marked as done:")
        for task in done_tasks:
            self.config.console.info(str(task))
        self.config.git.commit_and_push("do")



import click

from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.TodoFileNotFoundError import TodoFileNotFoundError


class Context(AbstractCommand):

    def run(self):
        try:
            click.echo(self())
        except TodoFileNotFoundError as e:
            e.echo_and_exit(self.config)

    def __call__(self, *args, **kwargs):
        self.taskManager.import_tasks()
        return self.taskManager.print_tree("contexts")


from pathlib import Path

import xdg_base_dirs
from todotree.Config.AbstractSubConfig import AbstractSubConfig


class Paths(AbstractSubConfig):
    """
    Configuration of the paths.
    Paths are calculated dynamically when not set.
    """

    def __init__(self):
        self.todo_folder: Path = xdg_base_dirs.xdg_data_home() / "todotree"
        """Path to the folder containing the data files."""
        self.__project_tree_folder: Path | None = None
        self.__todo_file: Path | None = None
        self.__done_file: Path | None = None
        self.__recur_file: Path | None = None
        self.__addons_folder: Path | None = None
        self.__stale_file: Path | None = None

    def read_from_dict(self, new_values: dict):
        self.todo_folder = Path(new_values.get('folder', self.todo_folder)).expanduser()

        if (todo_file := new_values.get('todo_file')) is not None:
            self.todo_file = Path(todo_file).expanduser()

        if (done_file := new_values.get('done_file')) is not None:
            self.done_file = Path(done_file).expanduser()

        if (recur_file := new_values.get('recur_file')) is not None:
            self.__recur_file = Path(recur_file).expanduser()

        if (stale_file := new_values.get('stale_file', None)) is not None:
            self.__stale_file = Path(stale_file).expanduser()

        if (project := new_values.get('project_folder')) is not None:
            self.__project_tree_folder = Path(project).expanduser()

        if (addons := new_values.get('addons_folder')) is not None:
            self.__addons_folder = Path(addons).expanduser()

    def apply_to_dict(self, dict_to_modify: dict):
        dict_to_modify['folder'] = str(self.todo_folder)
        dict_to_modify['todo_file'] = str(self.todo_file)
        dict_to_modify['done_file'] = str(self.done_file)
        dict_to_modify['recur_file'] = str(self.recur_file)
        dict_to_modify['stale_file'] = str(self.stale_file)
        dict_to_modify['project_folder'] = str(self.project_tree_folder)
        dict_to_modify['addons_folder'] = str(self.addons_folder)

    def __repr__(self):
        return (f"Paths({self.todo_folder}, "
                f"{self.todo_file}, {self.done_file}, {self.recur_file}, {self.addons_folder})")

    @property
    def todo_file(self) -> Path:
        """Path to the todo.txt file."""
        return self.__todo_file if self.__todo_file is not None else Path(self.todo_folder) / "todo.txt"

    @todo_file.setter
    def todo_file(self, new_path: Path):
        self.__todo_file = new_path

    @property
    def done_file(self) -> Path:
        """Path to the done.txt file."""
        return self.__done_file if self.__done_file is not None else Path(self.todo_folder) / "done.txt"

    @done_file.setter
    def done_file(self, new_path: Path):
        self.__done_file = new_path

    @property
    def recur_file(self) -> Path:
        """Path to the recur.txt file."""
        return self.__recur_file if self.__recur_file is not None else Path(self.todo_folder) / "recur.txt"

    @recur_file.setter
    def recur_file(self, new_recur_file: Path):
        self.__recur_file = new_recur_file

    @property
    def stale_file(self) -> Path:
        """Path to the stale.txt file."""
        return self.__stale_file if self.__stale_file is not None else Path(self.todo_folder) / "stale.txt"

    @property
    def addons_folder(self) -> Path:
        """Path to the addon folder."""
        return self.__addons_folder if self.__addons_folder is not None else Path(self.todo_folder) / "addons"

    @addons_folder.setter
    def addons_folder(self, new_path: Path):
        self.__addons_folder = new_path

    @property
    def project_tree_folder(self) -> Path:
        """Path to the folder containing the projects."""
        return self.__project_tree_folder if self.__project_tree_folder is not None else self.todo_folder / "projects"

    @project_tree_folder.setter
    def project_tree_folder(self, new_path: Path):
        self.__project_tree_folder = new_path



from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.NoSuchTaskError import NoSuchTaskError
from todotree.Errors.StaleFileNotFoundError import StaleFileNotFoundError
from todotree.Managers.StaleManager import StaleManager
from todotree.Task.Task import Task


class StaleRevive(AbstractCommand):

    def run(self, task_number: int):
        try:
            result = self(task_number)
        except StaleFileNotFoundError as e:
            e.echo_and_exit(self.config)
            exit(1) # IDE hinting.
        except NoSuchTaskError as e:
            self.config.console.error(e.message)
            exit(1)

        self.config.console.info("Revived the following task:")
        for task in result:
            self.config.console.info(task)
        self.config.git.commit_and_push("Stale Revive")

    def __call__(self, task_number: int) -> list[Task]:
        stale_manager = StaleManager(self.config)
        stale_manager.import_tasks()
        revived_task = stale_manager.remove_task_from_file(task_number)
        self.taskManager.import_tasks()
        return self.taskManager.add_tasks_to_file([revived_task])


import click
from todotree.Errors.TodotreeError import TodotreeError

from todotree.Commands.AbstractCommand import AbstractCommand


class Edit(AbstractCommand):
    def run(self, filename: str):
        # Disable fancy imports.
        self.config.enable_project_folder = False
        # Run based on which file is actually there.
        match filename:
            case "todo":
                click.edit(filename=str(self.config.paths.todo_file))
            case "done":
                click.edit(filename=str(self.config.paths.done_file))
            case "stale":
                click.edit(filename=str(self.config.paths.stale_file))
            case "recur":
                click.edit(filename=str(self.config.paths.recur_file))
            case _:
                raise TodotreeError(f"{filename} is not a correct file name to edit.")
        # Push the results to Git.
        self.config.git.commit_and_push("edit")

    def __call__(self, *args, **kwargs):
        raise NotImplemented


from todotree.Config.AbstractSubConfig import AbstractSubConfig


class TreePrint(AbstractSubConfig):
    """
    Class containing the prefixes for printing trees.
    """
    def __init__(self, t_print: str = "t", l_print: str = "l", s_print: str = "s", e_print: str = "e"):
        """
        Initialize a new instance of the TreePrint class.
        @param t_print: A string representing the "t" piece to print.
        @param l_print: A string representing the "l" piece to print.
        @param s_print: A string representing the "s" piece to print.
        @param e_print: A string representing the "e" piece to print.
        """
        self.e_print = e_print
        self.s_print = s_print
        self.l_print = l_print
        self.t_print = t_print

    def apply_to_dict(self, dict_to_modify: dict):
        """
        Modifies the dict object with new values.
        @param dict_to_modify: the dict object to modify.
        """
        dict_to_modify['t'] = self.t_print
        dict_to_modify['l'] = self.l_print
        dict_to_modify['s'] = self.s_print
        dict_to_modify['e'] = self.e_print

    def read_from_dict(self, new_values: dict):
        """Reads from a dict and applies new values."""
        self.t_print = new_values.get('t', self.t_print)
        self.l_print = new_values.get('l', self.l_print)
        self.s_print = new_values.get('s', self.s_print)
        self.e_print = new_values.get('e', self.e_print)

    def __repr__(self):
        return f"TreePrint({self.t_print}, {self.l_print}, {self.s_print}, {self.e_print})"


from todotree.Config.Config import Config

from todotree.Errors.TodotreeError import TodotreeError


class StaleFileNotFoundError(TodotreeError):
    """Represents an error indicating that the stale.txt file is not found."""

    def echo_and_exit(self, config: Config):
        config.console.error("The stale.txt file could not be found.")
        config.console.error(f"It searched at the following location: {config.paths.stale_file}")
        config.console.verbose(str(self))
        exit(1)



import click

from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.TodoFileNotFoundError import TodoFileNotFoundError


class Due(AbstractCommand):
    def __call__(self):
        # Disable fancy imports, because they do not have due dates.
        self.config.enable_project_folder = False
        self.taskManager.import_tasks()
        return self.taskManager.print_tree("due")

    def run(self):
        try:
            click.echo(self())
        except TodoFileNotFoundError as e:
            e.echo_and_exit(self.config)



from todotree.Errors.TaskParseError import TaskParseError
from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.TodoFileNotFoundError import TodoFileNotFoundError
from todotree.Task.Task import Task


class Add(AbstractCommand):

    def run(self, task: tuple):
        # Convert tuple to string
        try:
            task = self(" ".join(map(str, task)))
            self.config.console.info("Task added:")
            self.config.console.info(str(task))
            self.config.git.commit_and_push("add")
        except TodoFileNotFoundError as e:
            e.echo_and_exit(self.config)

    def __call__(self, task: str) -> Task:
        """Adds the new task to taskmanager and return the new task."""
        if not task:
            raise TaskParseError("Task is empty!")
        self.config.enable_project_folder = False
        self.taskManager.import_tasks()
        return self.taskManager.add_tasks_to_file([Task(0, task)])[0]



from collections import OrderedDict
from typing import Callable

from todotree.Config.TreePrint import TreePrint
from todotree.Task.Task import Task, DueDateBand


class Tree:
    """
    Class representing a tree.

    The tree is a graph structure where there are no loops.
    However, this one is a special version of it.
    This tree is only 2 nodes deep.
    The leaves are the actual objects.
    The first layer of nodes consist of the value of one of the attributes of the leaves.
    Leaves with the same attribute are connected to the same node.

    To print the tree, simply cast it to string, like so `str(Tree())`.
    """

    def __init__(self, list_to_convert, key, **kwargs):
        """
        :param list_to_convert: the list to convert to a tree.
        :param key: the key of the first node in the tree.

        :param Optional arguments:
          - root_name: The display name of the root node.
          - empty_string: The display name of the "empty" or "default" first node.
          - config: The configuration file for printing the tree.
        """

        self.data_structure: OrderedDict = OrderedDict()
        """
        The data structure of this tree.
        """

        self.root_name: str = kwargs.pop("root_name", "root")
        """
        The name of the root of the tree.
        """

        self.empty_string: str = kwargs.pop("empty_string", "default")
        """
        The 'empty' category.
        """

        self.key = key

        self.treeprint: TreePrint = kwargs.pop('treeprint', TreePrint("t", "l", "s", "e"))
        """
        The application configuration.
        """

        self.print_func: Callable[[str, Task], str] = kwargs.pop('print_func', lambda key_value, task: str(task))
        """
        Function which prints the task.
        Input are: key_value: The value of the key given by `key`.
        Output must be a string.
        """

        # Generate tree.
        self.generate_tree(list_to_convert, key)

    def generate_tree(self, list_to_parse: list, key: str):
        """
        Generate a tree like dictionary from the task list.
        The first line is each distinct key.
        Which points to a list of tasks. Tasks can appear in multiple lists.

        :list list_to_parse: The list to parse.
        :key: The key to index to, such as due:date, project or context.
        """
        for task in list_to_parse:
            if isinstance(getattr(task, key), list) and len(getattr(task, key)) > 0:
                for key_value in getattr(task, key):
                    # Also sets a default value if the key value does not exist.
                    self.data_structure.setdefault(key_value, list()).append(task)
            else:
                # It is a single item.
                self.data_structure.setdefault(
                    getattr(task, key) if getattr(task, key) else self.empty_string,
                    list()
                ).append(task)

    def __str__(self):
        """
        Used for pretty printing the tree.
        """
        # Edge case: empty list -> empty string.
        if self.data_structure == {}:
            return ""

        return self.__outer_loop(self.root_name + "\n")

    def sort(self):
        """
        Sort the tree.

        The outer leaves are sorted by task priority, then task number.
        The inner nodes are sorted by project/context, with the no_project as the first node.
        :returns: itself for function chaining.
        """
        # Sort the outer leaves.
        for inner in self.data_structure.values():
            inner.sort()
        if self.key == "due_date_band":
            # We do a different sorting for due, because that one has a
            # fixed sorting.
            for key in DueDateBand:
                if self.data_structure.get(key):
                    self.data_structure.move_to_end(key)
            return self
        # Sort the inner leaves by creating a new data structure.
        # The inner leaves are sorted by the most import job in their own tree.
        keys = [(k, v[0]) for k, v in self.data_structure.items()]
        keys.sort(key=lambda x: x[1])
        for key, _ in keys:
            self.data_structure.move_to_end(key)
        # Set the empty key to the front, if it has any tasks.
        if self.data_structure.get(self.empty_string):
            self.data_structure.move_to_end(self.empty_string, last=False)
        return self

    def __outer_loop(self, return_string):
        """Construct the string over the dict."""
        for i, first_node in enumerate(self.data_structure.keys()):
            # Print the outer node. This prints the value of the key.
            return_string += self.treeprint.t_print if (i < len(self.data_structure) - 1) else self.treeprint.s_print
            # Section with the name of the first node.
            return_string += first_node + "\n"
            # Inner loop.
            return_string = self.__inner_loop(self.data_structure[first_node], first_node, i, len(self.data_structure),
                                              return_string)
        return return_string

    def __inner_loop(self, first_list, first_node, i, len_first_node, return_string):
        """Construct the string over the list."""
        for j, second_node in enumerate(first_list):
            return_string += (
                    (self.treeprint.l_print if (i < len_first_node - 1) else self.treeprint.e_print)  # First prefix.
                    + (self.treeprint.t_print if (j < len(first_list) - 1) else self.treeprint.s_print)  # 2nd prefix.
                    + self.print_func(first_node, second_node) + "\n")  # Actual task.
        return return_string



import math
import shutil
from abc import ABC
from pathlib import Path
# https://docs.python.org/3/library/tempfile.html
from tempfile import NamedTemporaryFile

from todotree.Errors.NoSuchTaskError import NoSuchTaskError
from todotree.Task.ConsoleTask import ConsoleTask
from todotree.Task.Task import Task


class AbstractManager(ABC):
    not_found_error = FileNotFoundError
    """Error to raise when the database file is not found."""

    @property
    def max_task_number(self) -> int:
        """Property which defines the highest number in the task list."""
        try:
            return max([x.i for x in self.task_list])
        except ValueError:
            # Then the list is empty
            return 0

    @property
    def number_of_digits(self) -> int:
        """Property which defines the number of digits of the task in the task list with the highest number."""
        return int(math.ceil(math.log(self.max_task_number + 1, 10)))

    def __init__(self):
        self.file: Path = Path("/")
        """Path to the 'database' file. Must be set in the subclass."""

        self.task_list: list[Task] = []
        """Task list"""

    def remove_task_from_file(self, task_number: int) -> Task:
        # Remove task.
        try:
            removed_task = self.task_list.pop(task_number - 1)
        except IndexError as e:
            raise NoSuchTaskError(f"Task {task_number} does not exist in {self.file.name}.") from e
        self.write_to_file()
        return removed_task

    def remove_tasks_from_file(self, task_numbers: list[int]) -> list[Task]:
        removed_tasks = []
        task_numbers.sort()
        task_numbers.reverse()
        for task_number in task_numbers:
            try:
                t = self.get_by_task_number(task_number)
                self.task_list.remove(t)
                removed_tasks.append(t)
            except IndexError as e:
                raise NoSuchTaskError(f"Task {task_number} does not exist in {self.file}.") from e
        self.write_to_file()
        return removed_tasks

    def write_to_file(self):
        """Writes the entire list to the file."""
        #  Delete=false is needed for windows, I hope that somebodies temp folder won't be clobbered with this...
        try:
            with NamedTemporaryFile("w+t", newline="", delete=False) as temp_file:
                # may strip new lines by using task list.
                for counter in range(1, self.max_task_number + 1):
                    try:
                        temp_file.write(self.get_by_task_number(counter).to_file())
                    except NoSuchTaskError:
                        temp_file.write("\n")
                temp_file.flush()
                shutil.copy(temp_file.name, self.file)
        except FileNotFoundError as e:
            raise self.not_found_error from e

    def get_task_by_task_number(self, task_number: int) -> Task | None:
        for task in self.task_list:
            if task.i == task_number:
                return task
        return None

    def add_tasks_to_file(self, tasks: list[Task]) -> list[Task]:
        """Adds multiple tasks to the file. Returns the added tasks."""
        if len(self.task_list) != 0:
            numbers = set(x.i for x in self.task_list)
            blanks = list(set(range(1, self.max_task_number)) - numbers)
            blanks.sort()
            if len(blanks) >= len(tasks):
                for task in tasks:
                    task.i = blanks.pop(0)
            else:
                # Fill in the blanks.
                i = 0 # Warning, is modified in the following loop.
                for i, blank in enumerate(blanks):
                    tasks[i].i = blank
                # Add the remaining to the end of the list.
                for j, task in enumerate(tasks[i:], start=1):
                    task.i = j + self.max_task_number
        else:
            # Empty file, no tasks yet.
            for j, task in enumerate(tasks, start=1):
                task.i = j

        # Final housekeeping.
        self.task_list.extend(tasks)
        self.write_to_file()
        return tasks

    def import_tasks(self):
        """Imports the tasks from the database file."""
        try:
            with self.file.open('r') as f:
                content = f.readlines()
                for i, task in enumerate(content, start=1):
                    # Skip empty lines.
                    if task.strip() == "":
                        continue
                    self.task_list.append(Task(i, task.strip()))
        except FileNotFoundError as e:
            raise self.not_found_error() from e

    def __str__(self):
        """List the tasks."""
        s = ""
        for tsk in [ConsoleTask(task.i, task.task_string, self.number_of_digits) for task in self.task_list]:
            s += str(tsk) + "\n"
        return s

    def get_by_task_number(self, num: int):
        """
        Gets a task with number :num:
        :raise: NoSuchTaskError if the task is not found.
        """
        for task in self.task_list:
            if task.i == num:
                return task
        raise NoSuchTaskError
    def __repr__(self):
        return f"AbstractManager({self.max_task_number} Tasks)"



from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.NoSuchTaskError import NoSuchTaskError
from todotree.Errors.TodoFileNotFoundError import TodoFileNotFoundError


class Append(AbstractCommand):
    def __call__(self, task_nr: int, append_string: str):
        # Disable fancy imports, because they are not needed.
        self.config.enable_project_folder = False
        self.taskManager.import_tasks()
        return self.taskManager.append_to_task(task_number=task_nr, task_string=append_string)

    def run(self, task_nr: int, append_string: tuple[str]):
        # Append task.
        try:
            new_task = self(task_nr, " ".join(append_string))
        except TodoFileNotFoundError as e:
            e.echo_and_exit(self.config)
        except NoSuchTaskError as e:
            self.config.console.error(e.message)
            exit(1)
        self.config.console.info("The new task is: ")
        self.config.console.info(new_task)
        self.config.git.commit_and_push("append")



import datetime
import re
from enum import Enum

from todotree.Errors.RecurParseError import RecurParseError
from todotree.Task.Task import Task


class RecurTask:
    recur_pattern = re.compile(r"^\s*(\d{4}-\d{2}-\d{2})\s*;"  # Start Date.
                               r"\s*(\d{4}-\d{2}-\d{2})?\s*;"  # End date. 
                               r"\s*(\w+)\s*-"  # interval.
                               r"\s*(.*)$")  # Task.

    class Recurrence(Enum):
        Never = 0
        Daily = 1
        Weekly = 2
        Monthly = 3
        Yearly = 4

    def __init__(self, recur_line: str):
        """Parses `recur_line` to create a RecurTask."""
        self.start_date = datetime.date(1, 1, 1)
        self.end_date = datetime.date(9999, 1, 1)
        self.interval: RecurTask.Recurrence = RecurTask.Recurrence.Never
        self.task_string: str = ""
        self.recur_line: str = recur_line
        if recur_line.strip():
            # Parse each part
            match = self.recur_pattern.match(recur_line)
            if not match:
                raise RecurParseError(f"This string does not match regex: {recur_line}")
            start, end, interval, task_string = match.groups()
            self.task_string = task_string.strip()
            self.__parse_start_date(start)
            self.__parse_end_date(end)
            self.__parse_interval(interval)

    def __repr__(self):
        return f"RecurTask({self.recur_line.strip()})"

    def to_task(self, i: int = -1) -> Task:
        """Returns a Task representation of the RecurTask."""
        return Task(i, self.task_string)

    def __parse_interval(self, interval):
        try:
            self.interval = RecurTask.Recurrence[interval.capitalize()]
        except KeyError as e:
            raise RecurParseError(f"Error parsing interval: {e}. "
                                  f"It must be one of the following {', '.join([x.name for x in RecurTask.Recurrence])}")

    def __parse_end_date(self, end):
        if end:
            try:
                self.end_date = datetime.date.fromisoformat(end)
            except ValueError as e:
                raise RecurParseError(f"Error parsing end date: {e}")

    def __parse_start_date(self, start):
        try:
            self.start_date = datetime.date.fromisoformat(start)
        except ValueError as e:
            raise RecurParseError(f"Error parsing start date: {e}")



from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path

from git import Repo, InvalidGitRepositoryError, NoSuchPathError, GitCommandError, FetchInfo
from git.util import IterableList

from todotree.Config.AbstractSubConfig import AbstractSubConfig
from todotree.Config.ConsolePrefixes import ConsolePrefixes
from todotree.Errors.GitError import GitError


class GitHandler(AbstractSubConfig):
    """
    Handles all git related functionality.
    """

    class GitModeEnum(Enum):
        disabled = 0
        """Git functionality is disabled."""
        local = 1
        """Add and commits automatically"""
        full = 2
        """Also pulls and pushes to a remote repo"""

    def __init__(self, new_git_mode: str, pull_time_minutes: int, todo_folder: Path, console: ConsolePrefixes):
        self.todo_folder = todo_folder
        self.console = console
        self.pull_time_minutes = pull_time_minutes
        self.git_mode = new_git_mode

    @property
    def git_mode(self):
        return self.__git_mode

    @git_mode.setter
    def git_mode(self, new_mode):
        try:
            self.__git_mode: GitHandler.GitModeEnum = GitHandler.GitModeEnum[new_mode.lower()]
        except KeyError:
            self.console.warning(f"The git mode '{new_mode}' is not valid.")
            self.console.warning("The options are " + str([x.name for x in self.GitModeEnum]))
            self.console.warning("Continuing with the mode 'disabled'.")
            self.__git_mode = GitHandler.GitModeEnum.disabled
        if self.__git_mode is not GitHandler.GitModeEnum.disabled:
            try:
                self.repo: Repo = Repo(self.todo_folder)
            except InvalidGitRepositoryError:
                self.console.error("Git repository is not initialized.")
                self.console.error(f"Run `git init -C {self.todo_folder}` to initialize the repository.")
                self.console.error("Or disable the git functionality by setting git.mode to disabled in config.yaml")
                raise GitError("Git repository is not Initialized.")
            except NoSuchPathError:
                self.console.error("Directory is not found.")

    def git_pull(self):
        """
        Runs `git pull` on the TodoTree folder.

        - Does not pull if the previous pull time recent.
        - Only pulls if git_mode = Full
        """
        if self.__git_mode != GitHandler.GitModeEnum.full:
            return
        # Check last pull time.
        # Repo does not have a function to access FETCH_HEAD,
        # So this is done manually.
        fetch_head = Path(self.repo.git_dir) / "FETCH_HEAD"

        if fetch_head.exists():
            # Then the repo has been pulled once in its lifetime.
            if (datetime.now() - datetime.fromtimestamp(fetch_head.stat().st_mtime) <
                    timedelta(minutes=self.pull_time_minutes)):
                # Then the repo is pulled fairly recently. Do not do anything.
                self.console.verbose(
                    f"Repo was pulled recently at {datetime.fromtimestamp(fetch_head.stat().st_mtime)}. Not pulling.")
                return
        else:
            self.console.info("Pulling git repo for the first time.")

        # Pull the repo
        self.console.info("Pulling latest changes.")
        try:
            Repo(self.todo_folder).remote()
        except ValueError as e:
            self.console.error("Error Pulling Changes. The remote likely does not exist.")
            self.console.error("Run `git remote add origin https://example.org` to add the remote.")
            self.console.error(
                "Or disable this feature by setting git.mode to either `local` or `disabled` in config.yaml")
            raise GitError("Error Pulling changes, remote does not exist.") from e

        try:
            pull_result: IterableList[FetchInfo] = self.repo.git.pull()
            self.console.info(pull_result)
        except GitCommandError:
            GitError(f"Git pull failed.").warn_and_continue(self.console)

    def git_init(self):
        """Runs git init in the `todo_folder`."""
        try:
            Repo(self.todo_folder).init()
        except Exception as e:
            self.console.error("Error running git init")
            self.console.error(f"{e}")

    def commit_and_push(self, action: str):
        """
        Commit and push the files (if configured to do so).

        :param action: The name of the action, such as list or add.
        """
        if self.__git_mode is GitHandler.GitModeEnum.disabled:
            return

        if self.repo.is_dirty():
            self._commit(action)
            self._push()
        else:
            # git repo is not dirty, we do not have to commit anything.
            self.console.info("Nothing changed, nothing to commit or push.")

    def _push(self):
        if self.__git_mode is GitHandler.GitModeEnum.full:
            # Git push.
            try:
                result = self.repo.remote().push()
                result.raise_if_error()
                self.console.info(f"Push successful: {result[0].summary}")
            except GitCommandError as e:
                GitError(f"Push failed. {e.stderr}").warn_and_continue(self.console)

    def _commit(self, action):
        self.repo.index.add('*')
        try:
            # .dot files are not added...
            from todotree.Managers.RecurManager import RecurManager  # Adding globally causes a circular dependency.
            self.repo.index.add(RecurManager.recur_timestamp_filename)
        except:
            # File does not exist / Feature is not used.
            pass

        # Git commit.
        time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        new_commit = self.repo.index.commit(message=time + " " + action)
        self.console.info(f"Commit added: [{new_commit.hexsha[0:7]}] {new_commit.message} ")
        self.console.info(f"{new_commit.stats.total['files']} file(s) changed, "
                          f"{new_commit.stats.total['insertions']} insertions(+) "
                          f"{new_commit.stats.total['deletions']} deletions(-).")

    def apply_to_dict(self, dict_to_modify: dict):
        dict_to_modify['mode'] = str(self.git_mode.name)
        dict_to_modify['pull_delay'] = self.pull_time_minutes

    def read_from_dict(self, new_values: dict):
        self.git_mode = new_values.get("mode", self.git_mode)
        self.pull_time_minutes = new_values.get("pull_delay", self.pull_time_minutes)

    def __repr__(self):
        return f"GitHandler({self.__git_mode}, {self.pull_time_minutes}, {self.todo_folder})"



import click

from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.DoneFileNotFoundError import DoneFileNotFoundError
from todotree.Task.DoneTask import DoneTask
from todotree.Task.Task import Task


class AddX(AbstractCommand):

    def run(self, task: tuple):
        done = DoneTask.task_to_done(Task(0, " ".join(map(str, task))))
        if not self.config.paths.done_file.exists():
            DoneFileNotFoundError("").echo_and_exit(self.config)
        with self.config.paths.done_file.open("a") as f:
            f.write(done.to_file())
        click.echo(done)

    def __call__(self, *args, **kwargs):
        done = DoneTask.task_to_done(Task(0, " ".join(map(str, args))))
        with self.config.paths.done_file.open("a") as f:
            f.write(done.to_file())
        return done


import click

from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.DoneFileNotFoundError import DoneFileNotFoundError
from todotree.Errors.NoSuchTaskError import NoSuchTaskError
from todotree.Errors.TodoFileNotFoundError import TodoFileNotFoundError
from todotree.Managers.DoneManager import DoneManager
from todotree.Task.Task import Task


class Revive(AbstractCommand):
    def run(self, done_number: int):
        try:
            new_tasks = self(done_number)
        except DoneFileNotFoundError as e:
            e.echo_and_exit(self.config)
            exit(1)  # IDE hinting.
        except NoSuchTaskError as e:
            self.config.console.error(e.message)
            exit(1)  # IDE hinting.
        except TodoFileNotFoundError as e:
            e.echo_and_exit(self.config)
            exit(1)  # IDE hinting.
        self.config.console.info("The newly revived tasks:")
        for task in new_tasks:
            self.config.console.info(str(task))
        self.config.git.commit_and_push("revive")

    def __call__(self, done_number: int) -> list[Task]:
        done_manager = DoneManager(self.config)
        done_manager.import_tasks()
        revived_task = done_manager.get_task_by_task_number(done_number)
        if not revived_task:
            raise NoSuchTaskError(f"DoneTask {done_number} is not found.")
        self.taskManager.import_tasks()
        return self.taskManager.add_tasks_to_file([Task(revived_task.i, revived_task.task_string)])


import importlib

from todotree.Commands.AbstractCommand import AbstractCommand


class Version(AbstractCommand):

    def run(self):
        from importlib.metadata import version
        version_number = version("todotree")
        if self.config.console.is_verbose():
            self.config.console.verbose(f"The version is {version_number}")
        elif self.config.console.is_quiet():
            self.config.console.warning(version_number)
        else:
            self.config.console.info(f"Version: {version_number}")

    def __call__(self, *args, **kwargs):
        NotImplemented


from todotree.Config.Config import Config
from todotree.Errors.TodotreeError import TodotreeError


class TodoFileNotFoundError(TodotreeError):
    """
    Represents an error indicating that the todo.txt file is not found.
    """
    def echo_and_exit(self, config: Config):
        config.console.error("The todo.txt could not be found.")
        config.console.error(f"It searched at the following location: {config.paths.todo_file}")
        config.console.verbose(str(self))
        exit(1)



from functools import wraps
from pathlib import Path

import click

from todotree.Errors.ConfigFileNotFoundError import ConfigFileNotFoundError
from todotree.Config.Config import Config


class MainUtils:
    """
    Class containing static methods which are useful in the main class.
    """

    @staticmethod
    def initialize(config, **kwargs):
        """
        Finalizes configuration and does some setup which is applicable for all commands.
        """
        # Parse options.
        MainUtils.parse_common_options(config, **kwargs)

        # Git pull if needed.
        config.git.git_pull()
        # Logging
        config.console.verbose(f"Read configuration from {config.config_file}")
        config.console.verbose(f"The todo file is supposed to be at {config.paths.todo_file}")
        config.console.verbose(f"The done file is supposed to be at {config.paths.done_file}")

    @staticmethod
    def common_options(function):
        """
        Wrapper that defines common click options.

        It should be used as a decorator: `@common_options`

        This function is used to support both
        `todotree cd --verbose` and `todotree --verbose cd`
        """

        @wraps(function)
        @click.option('--config-file', default=None, help="Path to the configuration file.")
        @click.option('--todo-file', default=None, help="Path to the todo.txt file, overrides --config.")
        @click.option('--done-file', default=None, help="Path to the done.txt file, overrides --config.")
        @click.option('--verbose', is_flag=True, help="Increases verbosity in messages.", is_eager=True)
        @click.option('--quiet', is_flag=True, help="Do not print messages, only output. Useful in scripts.",
                      is_eager=True)
        @click.pass_context
        def wrapper(*args, **kwargs):
            return function(*args, **kwargs)

        return wrapper

    @staticmethod
    def parse_common_options(config: Config, **kwargs):
        """
        Parses the options from `common_options`.
        @param config: current config.
        @param kwargs: new options.
            :param config_file: `Optional[Path]` Path to the configuration file from the command line.
            :param done_file: `Optional[Path]` Path to the configuration file from the command line.
            :param todo_file: `Optional[Path]` Path to the configuration file from the command line.
            :param quiet: --quiet value
            :param verbose: --verbose value.
        """
        try:
            config.read(kwargs["config_file"])
        except ConfigFileNotFoundError as e:
            e.echo_and_exit(Config(), kwargs["config_file"], kwargs["verbose"])
        if kwargs["verbose"] and not config.console.is_quiet():
            config.console.set_verbose()
        if kwargs["quiet"]:
            config.console.set_quiet()
        if kwargs["todo_file"] is not None:
            config.paths.todo_file = Path(kwargs["todo_file"])
        if kwargs["done_file"] is not None:
            config.paths.done_file = Path(kwargs["done_file"])



import click

from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.TodoFileNotFoundError import TodoFileNotFoundError


class PrintRaw(AbstractCommand):

    def run(self):
        try:
            click.echo(self.taskManager.config.paths.todo_file.read_text())
        except FileNotFoundError:
            TodoFileNotFoundError("").echo_and_exit(self.config)

    def __call__(self, *args, **kwargs):
        raise NotImplemented


from todotree.Config.ConsolePrefixes import ConsolePrefixes
from todotree.Errors.TodotreeError import TodotreeError


class GitError(TodotreeError):
    """
    Represents an error indicating that something is wrong with the git repo.
    """
    def warn_and_continue(self, console: ConsolePrefixes):
        console.warning("An error occurred while trying to git pull.")
        console.warning("If this was unexpected, try again with the --verbose flag.")





from todotree.Config.Config import Config
from todotree.Errors.TodotreeError import TodotreeError


class DoneFileNotFoundError(TodotreeError):
    """
    Represents an error indicating that the done.txt file is not found.
    """
    def echo_and_exit(self, config: Config):
        config.console.error("The done.txt could not be found.")
        config.console.error(f"It searched at the following location: {config.paths.done_file}")
        config.console.verbose(self)
        exit(1)



from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.NoSuchTaskError import NoSuchTaskError
from todotree.Errors.StaleFileNotFoundError import StaleFileNotFoundError
from todotree.Managers.StaleManager import StaleManager
from todotree.Task.Task import Task


class StaleAdd(AbstractCommand):

    def run(self, task_number: int):
        try:
            task_to_add = self(task_number)
        except StaleFileNotFoundError as e:
            e.echo_and_exit(self.config)
            exit(1)  # IDE hinting.
        except NoSuchTaskError as e:
            self.config.console.error(e.message)
            exit(1)
        self.config.console.info("Added the following tasks to the stale list.")
        for task in task_to_add:
            self.config.console.info(str(task))
        self.config.git.commit_and_push("Stale Add")

    def __call__(self, task_number: int) -> list[Task]:
        stale_manager = StaleManager(self.config)
        stale_manager.import_tasks()
        self.taskManager.import_tasks()
        task_to_add = self.taskManager.remove_task_from_file(task_number)
        return stale_manager.add_tasks_to_file([task_to_add])


import datetime
import re
from datetime import datetime, date, timedelta
from enum import EnumType, Enum, StrEnum

from dateutil.utils import today

from todotree.Errors.InvalidPriorityError import InvalidPriorityError
from todotree.Errors.TaskParseError import TaskParseError


class DueDateBand(StrEnum):
    a = "Overdue"
    b = "Due today"
    c = "Due tomorrow"
    d = "Due in the next 7 days"
    e = "Due date further than the next 7 days"
    z = "No due date"


class Task:
    """
    Base class that represents a task.
    """

    t_date_pattern = re.compile(r"t:(\d{4})-(\d{2})-(\d{2})\s?")
    """Regex pattern which defines t:dates in a task string."""

    project_pattern = re.compile(r"\+\S+\s?")
    """Regex pattern which defines projects in a task string."""

    context_pattern = re.compile(r"@\w+\s?")
    """Regex pattern which defines contexts in a task string."""

    priority_pattern = re.compile(r"^\([a-zA-Z]\)\s?")
    """Regex pattern which defines the priority in a task string."""

    due_date_pattern = re.compile(r"due:(\d{4})-(\d{2})-(\d{2})\s?")
    """Regex pattern which defines due dates in a string.
    
    The format is due:yyyy-mm-dd.
    This is a due date for a single time.
    """

    due_month_pattern = re.compile(r"duy:(\d{2})-(\d{2})\s?")
    """Regex pattern which defines dum dates in a string.

    The format is due:mm-dd.
    This is a due date which recurs each year.
    
    An example case would be doing taxes. Each year you have to them before a certain date.
    """

    due_day_pattern = re.compile(r"dum:(\d{2})\s?")
    """Regex pattern which defines dum dates in a string.

    The format is due:dd.
    This is a due date for a deadline which recurs each month.
    """

    block_pattern = re.compile(r"(bl:)(\w+)\s?")
    """Regex pattern which defines a block key. format: bl:string"""

    blocked_by_pattern = re.compile(r"(by:)(\w+)\s?")
    """Regex pattern which defines a blocked by key. format: by:string"""

    @property
    def t_date(self) -> date | None:
        """Property which returns the latest t date of the task."""
        return max(self.t_date_all) if len(self.t_date_all) > 0 else None

    # FUTURE: t_date Setter.

    @property
    def due_date(self) -> date | None:
        """Property which returns the earliest due date."""
        return min(self.due_date_all) if len(self.due_date_all) > 0 else None

    # FUTURE: due_date setter. https://mathspp.com/blog/pydonts/properties

    @property
    def due_date_band(self) -> DueDateBand:
        """
        Property which returns the time until the due date lapses in human-readable language.
        """
        if self.due_date is None:
            return DueDateBand.z
        # List with options.
        difference = self.due_date - datetime.today().date()
        if difference < timedelta(days=0):
            return DueDateBand.a
        if difference < timedelta(days=1):
            return DueDateBand.b
        if difference < timedelta(days=2):
            return DueDateBand.c
        if difference < timedelta(days=7):
            return DueDateBand.d
        return DueDateBand.e

    @property
    def priority(self) -> int:
        """
        Property which defines the priority of a `Task` as an integer.
        """
        return ord(self.priority_string.upper()) - 65 if self.priority_string != "" else 684

    def __init__(self, i: int, task_string: str):
        """
        Initializes a single task from a string.
        :param i: The task identifier.
        :param task_string: The string containing the task.
        """
        self.date_marked_done: str | None = None
        """Date that this task was marked as done."""

        self.t_date_all: list[datetime.date] = []
        """All t dates of a task."""

        self.due_date_all: list[datetime.date] = []
        """All due dates of a task"""

        self.task_string: str = ""
        """raw un formatted string"""

        self.other_string: str = task_string
        """String containing any part which is not captured in another variable."""

        self.i: int = i
        """The task identifier"""

        self.priority_string: str = ""  # FUTURE: Convert to property instead.
        """priority as a String."""

        self.projects: list[str] = []
        """List of all projects"""

        self.contexts: list[str] = []
        """List of all contexts"""

        self.blocks: list[str] = []
        """List of all block identifiers."""

        self.blocked: list[str] = []
        """List of all blocked by identifiers"""

        # Parse the various properties from the task.
        if task_string:  # task is not empty
            self.task_string = task_string.strip()  # Remove whitespace and the \n.
            self.write_string = task_string

            # Parse the various items.
            self.__parse_priority(task_string)

            self.__parse_project(task_string)
            self.__parse_context(task_string)
            self.__parse_t_date(task_string)
            self.__parse_due_date(task_string)
            self.__parse_due_month(task_string)
            self.__parse_due_day(task_string)
            self.__parse_block_list(task_string)
            self.__parse_blocked_by_list(task_string)

            # Final cleanup of other_string.
            self.other_string = self.other_string.strip()

    def to_file(self):
        return self.task_string + "\n"

    def __parse_priority(self, task_string):
        """
        Parse the priority from the `task_string`.
        """
        if self.priority_pattern.match(task_string):
            with_parentheses = self.priority_pattern.match(task_string).group(0)
            self.priority_string = with_parentheses[1]
            self.other_string = self.priority_pattern.sub("", self.other_string).lstrip()

    def __parse_project(self, task_string):
        """
        Parse the projects from the `task_string`.
        """
        if self.project_pattern.search(task_string):
            projects_with_plus = self.project_pattern.findall(task_string)
            for p in projects_with_plus:
                # strip is for the trailing whitespace.
                self.projects.append(re.sub(r"\+", "", p).strip())
        self.other_string = self.project_pattern.sub("", self.other_string)

    def __parse_context(self, task_string):
        """
        Parse the contexts from the `task_string`.
        """
        if self.context_pattern.search(task_string):
            context_with_at = self.context_pattern.findall(task_string)
            for c in context_with_at:
                self.contexts.append(re.sub(r"@", "", c).strip())
        self.other_string = self.context_pattern.sub("", self.other_string)

    def __parse_t_date(self, task_string):
        """
        Parse the t_dates from the `task_string`.
        """
        for year, month, day in self.t_date_pattern.findall(task_string):
            try:
                # Add t:date to the t date list.
                self.t_date_all.append(datetime(int(year), int(month), int(day)).date())
            except ValueError:
                raise TaskParseError(f"This task has an incorrect t:date.{year}-{month}-{day}, task: {task_string}")
        self.other_string = self.t_date_pattern.sub("", self.other_string)

    def __parse_due_date(self, task_string):
        """
        Parse the due date from the `task_string`.
        """
        for year, month, day in self.due_date_pattern.findall(task_string):
            try:
                self.due_date_all.append(datetime(int(year), int(month), int(day)).date())
            except ValueError:
                raise TaskParseError(
                    f"This task has an incorrect due:date. date: {year}-{month}-{day}, task: {task_string}")
        self.other_string = self.due_date_pattern.sub("", self.other_string)

    def __parse_due_month(self, task_string):
        """
        Parse the duy date from the `task_string`.
        """
        for month, day in self.due_month_pattern.findall(task_string):
            try:
                date_next_year = datetime(datetime.today().year + 1, int(month), int(day))
                if date_next_year - today() < timedelta(weeks=4):
                    self.due_date_all.append(date_next_year.date())
                else:
                    self.due_date_all.append(datetime(datetime.today().year, int(month), int(day)).date())
            except ValueError:
                raise TaskParseError(f"This task has an incorrect dum:date. date: {month} {day}, task: {task_string}")
        self.other_string = self.due_month_pattern.sub("", self.other_string)

    def __parse_due_day(self, task_string):
        """
        Parses the dum date from `task_string`.
        """
        for day in self.due_day_pattern.findall(task_string):
            try:
                # Also take care of rollover near Christmas.
                date_next_month = datetime(
                    datetime.today().year if datetime.today().month < 12 else datetime.today().year + 1,
                    datetime.today().month + 1 if (datetime.today().month + 1) <= 12 else 1,
                    int(day))
                if date_next_month - today() < timedelta(days=7):
                    # The deadline in the next month is less than 7 days away.
                    self.due_date_all.append(date_next_month.date())
                else:
                    self.due_date_all.append(datetime(datetime.today().year, datetime.today().month, int(day)).date())
            except ValueError:
                raise TaskParseError(f"This task has an incorrect duy:date : {task_string}")
        self.other_string = self.due_day_pattern.sub("", self.other_string)

    def __parse_blocked_by_list(self, task_string):
        """
        Parse the blocked by list.
        """
        for blocked_item in self.blocked_by_pattern.finditer(task_string):
            self.blocked.append(str(blocked_item.group(2)))
        self.other_string = self.blocked_by_pattern.sub("", self.other_string)

    def __parse_block_list(self, task_string):
        """
        Parse the block list.
        """
        for item in self.block_pattern.finditer(task_string):
            self.blocks.append(str(item.group(2)))
        self.other_string = self.block_pattern.sub("", self.other_string)

    def add_or_update_priority(self, new_priority: str) -> bool:
        """
        Adds or updates the priority.
        :param new_priority: the new priority
        :return: A value indicating whether it was added or updated.
        True means it was added. False that it was updated.
        """
        if len(new_priority) > 1:
            # Then the string is not a single character.
            raise InvalidPriorityError(f"{new_priority} is not a valid priority.")
        if self.priority < 600:
            # Then there is already a priority defined.
            self.priority_string = new_priority
            self.task_string = self.priority_pattern.sub("(" + new_priority + ") ", self.task_string)
            return False
        else:
            # There was no priority defined.
            self.priority_string = new_priority
            self.task_string = "(" + new_priority + ") " + self.task_string
            return True

    def add_or_update_due(self, new_due: datetime) -> bool:
        """
        Updates the due date, or adds it if it does not exist.
        :param new_due: the new due date.
        :return: True if it was added. False if the due date is updated.
        """
        if self.due_date is not None:
            # There already are due dates, overwrite any other due dates.
            self.task_string = self.due_date_pattern.sub(
                "due:" + new_due.strftime("%Y-%m-%d"), self.task_string
            )
            self.due_date_all = [new_due.date()]
            return False
        # There was no due_date, substitute the default.
        self.task_string += " due:" + new_due.strftime("%Y-%m-%d")
        self.due_date_all.append(new_due.date())
        return True

    def add_or_update_t_date(self, new_t_date: str):
        """
        Adds or updates the t:date
        :param new_t_date: The new `t_date` in the format yyyy-mm-dd.
        """
        if self.t_date is not None:
            self.task_string = self.t_date_pattern.sub("t:" + new_t_date, self.task_string)
        else:
            self.task_string += " t:" + new_t_date
        self.t_date_all = [datetime.strptime(new_t_date, "%Y-%m-%d").date()]

    def __add__(self, other):
        """
        :param other: the string or task to append to this task.
        """
        match other:
            case str():
                return Task(self.i, self.task_string + " " + other.strip())
            case Task():
                return Task(self.i, self.task_string + " " + other.task_string)
            case _:
                return NotImplemented

    def __str__(self):
        prefix = self.i if self.i > 0 else ""
        task = self.task_string
        return str(prefix) + " " + task

    def __repr__(self):
        return f"Task({self.i}, {self.task_string})"

    def __lt__(self, other) -> bool:
        if not isinstance(other, Task):
            return NotImplemented
        # Sort by priority.
        if self.priority != other.priority:
            # Return < or > version.
            return self.priority < other.priority
        # Priority is the same, sort by task number.
        return self.i <= other.i



from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.DoneFileNotFoundError import DoneFileNotFoundError
from todotree.Managers.DoneManager import DoneManager


class ListDone(AbstractCommand):
    def run(self):
        try:
            self.config.console.info(self())
        except DoneFileNotFoundError as e:
            e.echo_and_exit(self.config)

    def __call__(self):
        done_manager = DoneManager(self.config)
        done_manager.import_tasks()
        return done_manager


from typing import override

from todotree.Errors.TodotreeError import TodotreeError


class RecurDateStampError(TodotreeError):
    """
    Represents an error when parsing the recur date stamp.
    """

    @override
    def echo_and_exit(self, config: 'Config', malformeddate: str):
        config.console.error(f"The date in the timestamp file is malformed {malformeddate}")
        config.console.error("You can repair this by adding the --date {date} option to recur.")
        config.console.verbose(str(self))
        exit(1)



from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.StaleFileNotFoundError import StaleFileNotFoundError
from todotree.Managers.StaleManager import StaleManager


class StaleList(AbstractCommand):

    def run(self):
        try:
            self.config.console.info(self())
        except StaleFileNotFoundError as e:
            e.echo_and_exit(self.config)

    def __call__(self):
        stale_manager = StaleManager(self.config)
        stale_manager.import_tasks()
        return str(stale_manager)


from todotree.Config.Config import Config
from todotree.Errors.StaleFileNotFoundError import StaleFileNotFoundError

from todotree.Managers.AbstractManager import AbstractManager
from todotree.Task.Task import Task


class StaleManager(AbstractManager):
    """Manages the stale.txt file"""
    not_found_error = StaleFileNotFoundError

    def __init__(self, configuration: Config):
        """Initializes a task manager, which manages the tasks."""

        super().__init__()
        self.task_list: list[Task] = []
        """List Without empty lines"""

        self.config: Config = configuration
        """Configuration settings for the application."""

        self.file = self.config.paths.stale_file
        """txt file location."""



from datetime import date

from todotree.Task.Task import Task


class DoneTask(Task):

    @staticmethod
    def task_to_done(tsk: Task) -> 'DoneTask':
        """
        Format the task String to a done state, by adding x 2020-02-02 in front of it.
        :param tsk: A task_string.
        :return: A done_string with a today's time stamp.
        """
        return DoneTask(tsk.i, "x " + str(date.today()) + " " + tsk.task_string)

    def to_file(self):
        """Returns a string useful for writing to a file."""
        return f"x {self.date_marked_done} {self.task_string}\n"

    @staticmethod
    def task_to_undone(tsk):
        """
        Removes the x 2020-02-02 part so the task can be added to the task file list
        :param tsk: A done string.
        :return: A task string.
        """
        # 13 is the number of chars of "x 2020-02-02 "
        return str.strip(tsk[13:])

    @staticmethod
    def chop(tsk):
        """Chops into a tuple containing the date and the task."""
        return str.strip(tsk[2:12]), str.strip(tsk[13:])

    def __init__(self, i: int, task_string: str):
        """
        Initializes a single task from a string.
        :param i: The task number.
        :param task_string: x 2020-02-02 The string containing the task.
        """
        date_marked_done, task_string = self.chop(task_string)
        super().__init__(i, task_string)
        self.date_marked_done = date_marked_done

    def __str__(self):
        return f"x {self.date_marked_done} {self.task_string}"

    def __repr__(self):
        return f"DoneTask(x {self.date_marked_done} {self.task_string})"




class TodotreeError(Exception):
    """
    Generic Exception class for application specific exceptions.
    """
    def __init__(self, message=""):
        self.message = message

        super().__init__(message)

    def __str__(self):
        return self.message

    def echo_and_exit(self, config: 'Config'):
        """
        Emit some error messages and exit the application.
        """
        config.console.error(str(self))
        exit(1)

    def warn_and_continue(self, config: 'Config'):
        """
        Emit some warning messages and continue the application.
        """
        config.console.warning(str(self))



from todotree.Errors.InvalidPriorityError import InvalidPriorityError
from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.TodoFileNotFoundError import TodoFileNotFoundError
from todotree.Task.Task import Task


class Priority(AbstractCommand):
    def run(self, task_number: int, new_priority: str):
        try:
            new_task = self(task_number, new_priority)
        except TodoFileNotFoundError as e:
            e.echo_and_exit(self.config)
        except InvalidPriorityError as e:
            e.echo_and_exit(self.config)
        self.config.console.info("The new task is: ")
        self.config.console.info(new_task)
        self.config.git.commit_and_push("priority")

    def __call__(self, task_number: int, new_priority: str):
        # Disable fancy imports.
        self.config.enable_project_folder = False
        self.taskManager.import_tasks()
        return self.taskManager.add_or_update_task(task_number, Task.add_or_update_priority, new_priority.upper())
